{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y-,Z1}
// Mairanda's EventNotify-like popups
// no dox! %-)
unit PopupsU;

interface

uses
  Windows, Messages;


const
  POPUP_NORMAL = 0;
  POPUP_WARNING = 1;
  POPUP_ERROR = 2;

  POPUP_OPACITY = 230;

  COLOR_WINDOW = 0;
  COLOR_TEXT = 1;
  COLOR_FR_UP = 2;
  COLOR_FR_DOWN = 3;

  PopupColors: array[0..2, 0..3] of DWORD =
  (($F7CFAD, $000000, $F7E7C6, $B58E6B),  // normal
   ($E7F7FF, $000000, $FFF7FF, $A5B6BD),  // warning
   ($0000BD, $E7F7FF, $1818D6, $00007B)); // error


type
  TPopupManager = class;

  TPopUp = class
  private
    FType: Integer;

    FX: Integer;
    FY: Integer;
    FWidth: Integer;
    FHeight: Integer;

    FName: string;
    FMessage: string;
    FTime: string;

    FWndH: DWORD;
    FIcon: DWORD;

    FNameExts: TSize;
    FTimeExts: TSize;
    FMsgExts: TSize;

    FTimerH: DWORD;
    FAlpha: Byte;

    FPhase: Integer;
    FPause: Integer;

    FMouseDowned: Integer;

    FPopMan: TPopupManager;

  public
    constructor Create(X, Y: Integer; Name, Msg: string; WType: Integer;
      Manager: TPopupManager; Pause: Integer; Invis: Boolean);
    destructor Destroy; override;

    function OnMouseDown(Button: Integer): Boolean; virtual;
    procedure Close(Immediate: Boolean);

    property PopupType: Integer read FType;
    property X: Integer read FX;
    property Y: Integer read FY;
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property WndH: DWORD read FWndH;
    property Icon: DWORD read FIcon;
  end;

  TPopupManager = class
  private
    FTop: Boolean;
    FLeft: Boolean;
    FPopupList: array of TPopup;

    FMinW: Integer;
    FMinH: Integer;

    FDefPause: Integer;
    FMaxMsgLen: Integer;

    FPopupCount: Integer;

    procedure DeletePopup(Popup: TPopup);

    procedure FindPopupCoords(var X, Y: Integer);

  public
    constructor Create(FromTop, FromLeft: Boolean; Pause: Integer);
    destructor Destroy; override;

    procedure Clear;
    procedure AddPopup(Name, Msg: string; WType: Integer);

    property MaxMsgLen: Integer read FMaxMsgLen;
    property PopupCount: Integer read FPopupCount;

    property FromTop: Boolean read FTop write FTop;
    property FromLeft: Boolean read FLeft write FLeft;
  end;


procedure InitPopups;
procedure DeinitPopups;


var
  PopupManager: TPopupManager = nil;
  ParentWindow: DWORD = 0;


implementation


const
  WS_EX_LAYERED = $00080000;
  LWA_ALPHA = $00000002;


var
  FontBoldH: DWORD = 0;
  FontH: DWORD = 0;
  BrushesH: array[0..2, 0..3] of DWORD;
  PensH: array[0..2, 0..3] of DWORD;

  Initialized: Boolean = False;
  Trans: Boolean = False;
  SetLayeredWindowAttributesProc:
    function(WndH: DWORD; crKey: TColorRef; bAlpha: Byte; dwFlags: DWORD): LongBool; stdcall = nil;


function SetLayeredWindowAttributes(WndH: DWORD; crKey: TColorRef; bAlpha: Byte;
  dwFlags: DWORD): LongBool;
begin
  if Assigned(SetLayeredWindowAttributesProc) then
    Result := SetLayeredWindowAttributesProc(WndH, crKey, bAlpha, dwFlags)
  else
    Result := False;
end;

procedure DrawPopup(PopObj: TPopUp; DC: DWORD);
var
  R: TRect;
  OldOH: DWORD;
begin
  if (PopObj = nil) or (DC = 0) then exit;

  R.Left := 1;
  R.Top := 1;
  R.Right := PopObj.Width - 1;
  R.Bottom := PopObj.Height - 1;
  FillRect(DC, R, BrushesH[PopObj.PopupType, COLOR_WINDOW]);
  MoveToEx(DC, 0, PopObj.FHeight, nil);
  //frame left and top
  OldOH := SelectObject(DC, PensH[PopObj.PopupType, COLOR_FR_UP]);
  LineTo(DC, 0, 0);
  LineTo(DC, PopObj.Width - 1, 0);
  SelectObject(DC, OldOH);
  // frame right and bottom
  OldOH := SelectObject(DC, PensH[PopObj.PopupType, COLOR_FR_DOWN]);
  LineTo(DC, PopObj.Width - 1, PopObj.Height - 1);
  LineTo(DC, 0, PopObj.Height - 1);
  SelectObject(DC, OldOH);
  // draw text
  SetBkMode(DC, TRANSPARENT);
  SetTextColor(DC, PopupColors[PopObj.PopupType, COLOR_TEXT]);
  // name
  OldOH := SelectObject(DC, FontBoldH);
  TextOut(DC, 3 + 16 + 3, 4, PChar(PopObj.FName), Length(PopObj.FName));
  SelectObject(DC, OldOH);
  // time
  OldOH := SelectObject(DC, FontH);
  TextOut(DC, PopObj.FWidth - 4 - PopObj.FTimeExts.CX, 6, PChar(PopObj.FTime), Length(PopObj.FTime));
  SelectObject(DC, OldOH);
  // message
  OldOH := SelectObject(DC, FontH);
  TextOut(DC, 3 + 16 + 3, 4 + 4 + PopObj.FNameExts.CY, PChar(PopObj.FMessage), Length(PopObj.FMessage));
  SelectObject(DC, OldOH);
  // icon
  DrawIconEx(DC, 3, 10, PopObj.FIcon, 16, 16, 0, 0, DI_NORMAL);
end;

function WndProc(WndH: HWND; Msg: UINT; wPar: WPARAM; lPar: LPARAM): LRESULT; stdcall;
var
  PopObj: TPopUp;
  R: TRect;
  PS: TPaintStruct;
  DoBP: Boolean;
  DC: DWORD;
  BN: Integer;
begin
  PopObj := TPopUp(GetWindowLong(WndH, GWL_USERDATA));
  case Msg of
    WM_ERASEBKGND: Result := 1;
    WM_PAINT:
      begin
        DoBP := GetUpdateRect(WndH, R, False);
        if DoBP then DC := BeginPaint(WndH, PS) else DC := 0;
        DrawPopup(PopObj, DC);
        if DoBP then EndPaint(WndH, PS);
        Result := 1;
      end;
    WM_CLOSE:
      begin
        if ((PopObj <> nil) and (PopObj.FPhase = 3)) or (PopObj = nil) then
          Result := DefWindowProc(WndH, Msg, wPar, lPar)
        else
          Result := 1;
      end;
    WM_DESTROY:
      begin
        if (PopObj <> nil) then
        begin
          SetWindowLong(WndH, GWL_USERDATA, 0);
          PopObj.Free;
        end;
        Result := DefWindowProc(WndH, Msg, wPar, lPar);
      end;
    WM_LBUTTONDOWN, WM_RBUTTONDOWN:
      begin
        if (Msg = WM_LBUTTONDOWN) then BN := 1 else BN := 2;

        if (PopObj <> nil) then
        begin
          if not PopObj.OnMouseDown(BN) then PopObj.Close(False);
        end
        else
          SendMessage(WndH, WM_CLOSE, 2, 0);
        Result := 0;
      end;
    WM_TIMER:
      begin
        if (PopObj <> nil) then
        begin
          if (wPar = 1) then
          begin
            if (PopObj.FAlpha >= POPUP_OPACITY) then
            begin
              PopObj.FAlpha := POPUP_OPACITY;
              SetLayeredWindowAttributes(WndH, 0, POPUP_OPACITY, LWA_ALPHA);
              KillTimer(WndH, PopObj.FTimerH);
              PopObj.FTimerH := 0;
              PopObj.FPhase := 1;
              PopObj.FTimerH := SetTimer(WndH, 2, PopObj.FPause, nil);
              if (PopObj.FTimerH = 0) then
              begin
                PopObj.FPhase := 3;
                PostMessage(WndH, WM_CLOSE, 0, 0);
              end;
            end
            else
            begin
              Inc(PopObj.FAlpha, 10);
              SetLayeredWindowAttributes(WndH, 0, PopObj.FAlpha, LWA_ALPHA);
            end;
          end
          else
          if (wPar = 2) then
          begin
            KillTimer(WndH, PopObj.FTimerH);
            PopObj.FPhase := 2;
            PopObj.FTimerH := SetTimer(WndH, 3, 50, nil);
            if (PopObj.FTimerH = 0) then
            begin
              PopObj.FPhase := 3;
              PostMessage(WndH, WM_CLOSE, 0, 0);
            end;
            if not Trans then PopObj.FAlpha := 0;
          end
          else
          if (wPar = 3) then
          begin
            if (PopObj.FAlpha <= 10) then
            begin
              PopObj.FAlpha := 0;
              SetLayeredWindowAttributes(WndH, 0, 0, LWA_ALPHA);
              KillTimer(WndH, PopObj.FTimerH);
              PopObj.FTimerH := 0;
              PopObj.FPhase := 3;
              PostMessage(WndH, WM_CLOSE, 0, 0);
            end
            else
            begin
              Dec(PopObj.FAlpha, 10);
              SetLayeredWindowAttributes(WndH, 0, PopObj.FAlpha, LWA_ALPHA);
            end;
          end
        end;
        Result := 0;
      end;
    else
      Result := DefWindowProc(WndH, Msg, wPar, lPar);
  end;
end;

procedure InitPopups;
var
  WndClass: TWndClass;
  HInst: DWORD;
  F, C: Integer;
begin
  if Initialized then exit;

  if (ParentWindow = 0) then
    ParentWindow := CreateWindow('#32770', 'K8 Popups parent window',
      WS_OVERLAPPEDWINDOW, 0, 0, 600, 450, 0, 0, hInstance, nil);

  HInst := GetModuleHandle(nil);
  ZeroMemory(@WndClass, SizeOf(WndClass));

  WndClass.Style := CS_DBLCLKS or CS_HREDRAW or CS_VREDRAW or CS_OWNDC or CS_SAVEBITS;
  WndClass.lpfnWndProc := @WndProc;
  WndClass.hInstance := HInst;
  WndClass.hCursor := LoadCursor(0, IDC_ARROW);
  WndClass.lpszClassName := 'K8_PopUp_Window_Class';
  WndClass.hbrBackground := COLOR_BTNFACE;
  RegisterClass(WndClass);

  FontH := CreateFont(
    8,
    0,  // width
    0,  // escapement
    0,  // orientation
    FW_DONTCARE, //FW_BOLD,  // weight FW_DONTCARE
    0,  // not italic
    0,  // not underlined
    0,  // not striked out
    ANSI_CHARSET,  // DEFAULT_CHARSET? OEM_CHARSET?
    OUT_DEFAULT_PRECIS,
    CLIP_DEFAULT_PRECIS,
    DRAFT_QUALITY,
    DEFAULT_PITCH or FF_SWISS,
    'MS Sans Serif');

  FontBoldH := CreateFont(
    8,
    0,  // width
    0,  // escapement
    0,  // orientation
    FW_BOLD, //FW_BOLD,  // weight FW_DONTCARE
    0,  // not italic
    0,  // not underlined
    0,  // not striked out
    ANSI_CHARSET,  // DEFAULT_CHARSET? OEM_CHARSET?
    OUT_DEFAULT_PRECIS,
    CLIP_DEFAULT_PRECIS,
    DRAFT_QUALITY,
    DEFAULT_PITCH or FF_SWISS,
    'MS Sans Serif');

  for F := 0 to 2 do
  begin
    for C := 0 to 3 do
    begin
      BrushesH[F, C] := CreateSolidBrush(PopupColors[F, C]);
      if (C = COLOR_FR_UP) or (C = COLOR_FR_DOWN) then
      begin
        PensH[F, C] := CreatePen(PS_SOLID, 0, PopupColors[F, C]);
      end;
    end;
  end;
  Initialized := True;
end;

procedure DeinitPopups;
var
  HInst: DWORD;
  F, C: Integer;
begin
  if not Initialized then exit;

  HInst := GetModuleHandle(nil);
  UnregisterClass('K8_PopUp_Window_Class', HInst);

  DeleteObject(FontH);
  DeleteObject(FontBoldH);
  FontH := 0;
  FontBoldH := 0;
  for F := 0 to 2 do
  begin
    for C := 0 to 3 do
    begin
      DeleteObject(BrushesH[F, C]);
      DeleteObject(PensH[F, C]);
    end;
  end;
  FillChar(BrushesH, SizeOf(BrushesH), 0);
  FillChar(PensH, SizeOf(PensH), 0);

  Initialized := False;
end;

function T2S(I: Integer): string;
begin
  Str(I:2, Result);
  if (Result[1] = ' ') then Result[1] := '0';
end;

constructor TPopup.Create(X, Y: Integer; Name, Msg: string; WType: Integer;
  Manager: TPopupManager; Pause: Integer; Invis: Boolean);
var
  WndStyle: DWORD;
  WndExStyle: DWORD;
  HInst: DWORD;
  DC: DWORD;
  OldFH: DWORD;
  STime: TSystemTime;
  W0, NewW, NewH: Integer;
  WorkArea: TRect;
begin
  if not Initialized then InitPopups;
  FPopMan := Manager;

  HInst := GetModuleHandle(nil);
  if not SystemParametersInfo(SPI_GETWORKAREA, 0, @WorkArea, 0) then
  begin
    WorkArea.Left := 0;
    WorkArea.Top := 0;
    WorkArea.Right := 640;
    WorkArea.Bottom := 480;
  end;

  FX := X;
  FY := Y;
  FWidth := 151;
  FHeight := 36;
  FName := Name;
  if (FName = '') then FName := ' ';
  FMessage := Msg;
  if (FMessage = '') then FMessage := ' ';
  GetLocalTime(STime);
  FTime := T2S(STime.wHour) + ':' + T2S(STime.wMinute);
  FType := WType;
  if (FType < 0) or (FType > POPUP_ERROR) then FType := POPUP_NORMAL;
  case WType of
    POPUP_NORMAL:
      begin
        FIcon := LoadIcon(0, IDI_ASTERISK);
      end;
    POPUP_WARNING:
      begin
        FIcon := LoadIcon(0, IDI_QUESTION);
      end;
    POPUP_ERROR:
      begin
        FIcon := LoadIcon(0, IDI_EXCLAMATION);
      end;
  end;

  WndStyle := WS_POPUP;
  WndExStyle := WS_EX_TOPMOST;
  if Trans then WndExStyle := WndExStyle or WS_EX_LAYERED;

  FWndH :=
    CreateWindowEx(WndExStyle, 'K8_PopUp_Window_Class', nil, WndStyle,
      FX, FY, FWidth, FHeight, ParentWindow,
      0, HInst, nil);
  if (FWndH = 0) then exit;
  SetWindowLong(FWndH, GWL_USERDATA, DWORD(Self));

  // determine extents
  DC := GetDC(FWndH);
  OldFH := SelectObject(DC, FontH);
  GetTextExtentPoint32(DC, PChar(FTime), Length(FTime), FTimeExts);
  GetTextExtentPoint32(DC, PChar(FMessage), Length(FMessage), FMsgExts);
  SelectObject(DC, OldFH);
  OldFH := SelectObject(DC, FontBoldH);
  GetTextExtentPoint32(DC, PChar(FName), Length(FName), FNameExts);
  SelectObject(DC, OldFH);
  ReleaseDC(FWndH, DC);

//  WriteLn('Name: ', FNameExts.CX, 'x', FNameExts.CY);
//  WriteLn('Time: ', FTimeExts.CX, 'x', FTimeExts.CY);
//  WriteLn('Msg: ', FMsgExts.CX, 'x', FMsgExts.CY);

  NewW := 3 + 16 + 4 + FNameExts.CX + 8 + FTimeExts.CX + 4;
  W0 := 3 + 16 + 4 + FMsgExts.CX + 4;
  if (W0 > NewW) then NewW := W0;
  if (FWidth < NewW) then FWidth := NewW;
  NewH := 4 + FNameExts.CY + 4 + FMsgExts.CY + 4;
  if (NewH > FHeight) then FHeight := NewH;
  if (FX + FWidth > WorkArea.Right) then FX := WorkArea.Right - FWidth;
  if (FY + FHeight > WorkArea.Bottom) then FX := WorkArea.Bottom - FHeight;
  SetWindowPos(FWndH, HWND_TOPMOST, FX, FY, FWidth, FHeight,
    SWP_NOACTIVATE or SWP_NOSENDCHANGING);

  FAlpha := POPUP_OPACITY;
  FPhase := 0;
  FTimerH := SetTimer(FWndH, 1, 50, nil);
  if (FTimerH <> 0) then
    FAlpha := 0
  else
  begin
    FPhase := 3;
    PostMessage(FWndH, WM_CLOSE, 0, 0);
  end;
  if Trans then SetLayeredWindowAttributes(FWndH, 0, FAlpha, LWA_ALPHA);

  FPause := Pause;
  if (FPause = 0) then FPause := MaxInt;
  if (FPause < 10) then FPause := 10;
  FMouseDowned := 0;

  if not Invis then ShowWindow(FWndH, SW_SHOWNOACTIVATE);
end;

destructor TPopup.Destroy;
begin
  DeleteObject(FIcon);
  KillTimer(FWndH, FTimerH);
  if (FPopMan <> nil) then FPopMan.DeletePopup(Self);
end;

function TPopup.OnMouseDown(Button: Integer): Boolean;
begin
  FMouseDowned := Button;
  Result := False;
end;

procedure TPopup.Close(Immediate: Boolean);
begin
  FPause := 1;
  if Immediate then FPhase := 3;
  if (FPhase = 1) then
    SendMessage(WndH, WM_TIMER, 2, 0)
  else
    SendMessage(WndH, WM_CLOSE, 2, 0);
end;

constructor TPopupManager.Create(FromTop, FromLeft: Boolean; Pause: Integer);
var
  P: TPopup;
begin
  FTop := FromTop;
  FLeft := FromLeft;
  SetLength(FPopupList, 0);
  P := TPopup.Create(0, 0, ' ', ' ', POPUP_NORMAL, nil, 10, True);
  FMinW := P.FWidth;
  FMinH := P.FHeight;
  DestroyWindow(P.FWndH);  // this will free P too

  FDefPause := Pause;
  FMaxMsgLen := 255;
  FPopupCount := 0;
end;

destructor TPopupManager.Destroy;
begin
  Clear;
  SetLength(FPopupList, 0);
end;

procedure TPopupManager.Clear;
var
  F: Integer;
begin
  for F := 0 to High(FPopupList) do
  begin
    if (FPopupList[F] <> nil) then
    begin
      FPopupList[F].FPopMan := nil;
      FPopupList[F].Close(True);
      FPopupList[F] := nil;
    end;
  end;
  SetLength(FPopupList, 0);
end;

procedure TPopupManager.FindPopupCoords(var X, Y: Integer);
var
  F: Integer;
  WorkArea: TRect;
begin
  if not SystemParametersInfo(SPI_GETWORKAREA, 0, @WorkArea, 0) then
  begin
    WorkArea.Left := 0;
    WorkArea.Top := 0;
    WorkArea.Right := 640;
    WorkArea.Bottom := 480;
  end;

  if FLeft then X := 0 else X := 32700;

  if FTop then
  begin
    // from top to bottom
    Y := WorkArea.Top - (FMinH + 4);
    for F := 0 to High(FPopupList) do
    begin
      if (FPopupList[F] <> nil) and (Y < FPopupList[F].FY) then Y := FPopupList[F].FY;
    end;
    Inc(Y, FMinH + 4);
    if (Y > WorkArea.Bottom - FMinH) then Y := WorkArea.Top;
  end
  else
  begin
    // from bottom to top
    Y := WorkArea.Bottom;
    for F := 0 to High(FPopupList) do
    begin
      if (FPopupList[F] <> nil) and (Y > FPopupList[F].FY) then Y := FPopupList[F].FY;
    end;
    Dec(Y, FMinH + 4);
    if (Y < WorkArea.Top) then Y := WorkArea.Bottom - FMinH - 4;
  end;
end;

procedure TPopupManager.AddPopup(Name, Msg: string; WType: Integer);
var
  P: TPopup;
  F: Integer;
  X, Y: Integer;
begin
  if (Length(Msg) > FMaxMsgLen) then Msg := Copy(Msg, 1, FMaxMsgLen - 3) + '...';
  FindPopupCoords(X, Y);
  P := TPopup.Create(X, Y, Name, Msg, WType, Self, FDefPause, False);
  Inc(FPopupCount);

  F := 0;
  while (F <= High(FPopupList)) and (FPopupList[F] <> nil) do Inc(F);

  if (F > High(FPopupList)) then
  begin
    SetLength(FPopupList, Length(FPopupList) + 3);
    FPopupList[F] := nil;
    FPopupList[F+1] := nil;
    FPopupList[F+2] := nil;
  end;
  FPopupList[F] := P;
end;

procedure TPopupManager.DeletePopup(Popup: TPopup);
var
  F: Integer;
begin
  FPopupCount := 0;
  F := 0;
  while (F <= High(FPopupList)) do
  begin
    if (FPopupList[F] = Popup) then FPopupList[F] := nil;
    if (FPopupList[F] <> nil) then Inc(FPopupCount);
    Inc(F);
  end;
end;


initialization
  FillChar(BrushesH, SizeOf(BrushesH), 0);
  FillChar(PensH, SizeOf(PensH), 0);
  SetLayeredWindowAttributesProc :=
    GetProcAddress(GetModuleHandle('user32.dll'), 'SetLayeredWindowAttributes');
  Trans := Assigned(SetLayeredWindowAttributesProc);

finalization
  PopupManager.Free;
  PopupManager := nil;
  DeinitPopups;
end.
