{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y-,Z1}
// :   SockU       
//   
{.$R-}
unit RefreshThreadU;

interface

uses
  Windows, Messages, SysUtils, Classes;

type
  TRefreshThread = class(TThread)
  protected
    CS: TRTLCriticalSection;  //    ...

    FRefreshOnline: Integer;//    - ( )
    FRefreshChat: Integer;  //     ( )
    FSayList: TStringList;  //    ,    

    //    ...
    FChatUpdateInterval: DWORD;
    FOnlineUpdateInterval: DWORD;

    FLogin: string;     //  ...
    FPassword: string;  //    

    FLastChatUpdate: DWORD;   //      ?
    FLastOnlineUpdate: DWORD; // ...  ?

    procedure Execute; override;  //  - 

    // 
    procedure Lock;
    procedure Unlock;

  public
    constructor Create(FormToNotifyH: DWORD; Login, Password: string;
      ChatUpdateInterval, OnlineUpdateInterval: Integer);
    destructor Destroy; override;

    procedure SetNewIntervals(ChatUpdateInterval, OnlineUpdateInterval: Integer);

    procedure Say(const Msg: string);
      //  , , ...  , -!
    procedure ForceOnlineRefresh;
      //  : "  - !"
    procedure ForceChatRefresh;
      //    : "  -!"
  end;


const
  WM_TAKEVENT = WM_USER + 13;  // wParam: event, lParam: data
  WM_TAKTHREADDIES = WM_USER + 14;

  TAKEVENT_LOGGING_IN = 1;
  TAKEVENT_LOGIN_FAILED = 2;
  TAKEVENT_LOGIN_OK = 3;

  TAKEVENT_LOGGING_OUT = 4;    //      
  TAKEVENT_LOGOUT_FAILED = 5;  //      
  TAKEVENT_LOGOUT_OK = 6;      //      

  TAKEVENT_REFRESHING_ONLINE = 7;
  TAKEVENT_ONLINE_FAILED = 8;  // lParam -- 
  TAKEVENT_ONLINE_OK = 9;  // lParam: TStringList

  TAKEVENT_REFRESHING_CHAT = 10;
  TAKEVENT_CHAT_FAILED = 11;  // lParam -- 
  TAKEVENT_CHAT_OK = 12;  // lParam: TStringList

  TAKEVENT_ARRIVES = 13;  // lParam: string-name
  TAKEVENT_LEAVES = 14;  // lParam: string-name (   )
  TAKEVENT_PERSONAL = 15;  // lParam: string-name#10message
  TAKEVENT_PRIVATE = 16;  // lParam: string-name#10message


implementation

uses
  SockU, LogU;


{ TRefreshThread }
constructor TRefreshThread.Create(FormToNotifyH: DWORD; Login, Password: string;
      ChatUpdateInterval, OnlineUpdateInterval: Integer);
begin
  InitializeCriticalSection(CS);

  SockU.FormToNotify := FormToNotifyH;
  FLogin := Login;
  FPassword := Password;

  FRefreshOnline := 0;
  FRefreshChat := 0;
  FSayList := TStringList.Create;

  FChatUpdateInterval := DWORD(ChatUpdateInterval) * 1000;
  FOnlineUpdateInterval := DWORD(OnlineUpdateInterval) * 60 * 1000;

  FLastChatUpdate := 0;
  FLastOnlineUpdate := 0;

  inherited Create(False);  //    ( -!)
end;

destructor TRefreshThread.Destroy;
begin
  //   ,   
  SendNotifyMessage(SockU.FormToNotify, WM_TAKTHREADDIES, 0, 0);

  FSayList.Free;  // -  ...
  DeleteCriticalSection(CS);
  inherited;
end;

procedure TRefreshThread.Execute;
var
  OL: TStringList;
  ToSay: string;
  DoUpd: Integer;
  Name, Msg: string;
begin
try
  FreeOnTerminate := True;  // - ...

  LogStr('receiving thread started');  // !

  //     ! ()
  SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_LOGGING_IN, 0);//  

  if (SockU.ChatId = '') then
  begin
    //    
    if not SockU.DoLogin(FLogin, FPassword) then
    begin
      // !  .    
      SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_LOGIN_FAILED, 0);
      exit;
    end;
  end
  else
    LogStr('login skipped');  // !

  //  
  SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_LOGIN_OK, 0);

  //     
  InterlockedExchange(FRefreshOnline, 1);
  InterlockedExchange(FRefreshChat, 1);

  // "<, !>    " ()...
  // ...   .
  repeat
    if Terminated then break;  //   ?

    //   ?
    DoUpd := InterlockedExchange(FRefreshOnline, 0);
    if (DoUpd <> 0) or (GetTickCount - FLastOnlineUpdate >= FOnlineUpdateInterval) then
    begin
      //        -  -
      LogStr('refreshing online list');  // 

      //   .    :
      //  ,     ...     .
      FLastOnlineUpdate := GetTickCount;

      //  
      SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_REFRESHING_ONLINE, 0);
      // ...
      OL := LoadAndParseOnlineList;
      // ...  
      if (OL = nil) then
        // .  
        SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_ONLINE_FAILED, 0)
      else
        //    
        SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_ONLINE_OK, Integer(OL));
    end;

    if Terminated then break;  //   ?

    // :     ?
    Lock;  //    
    if (FSayList.Count > 0) then
    begin
      //    (   ,  )
      ToSay := FSayList[0];
      FSayList.Delete(0);
    end
    else
      ToSay := '';
    Unlock;  //    

    //   ?
    DoUpd := InterlockedExchange(FRefreshChat, 0);
    if (ToSay <> '') or (DoUpd <> 0) or
       (GetTickCount - FLastChatUpdate >= FChatUpdateInterval) then
    begin
      //     ,      -  
      LogStr('refreshing chat');  // 

      //   .    :
      //  ,     ...     .
      FLastChatUpdate := GetTickCount;

      //  
      SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_REFRESHING_CHAT, 0);
      //     ( )
      ArrivedNames := '';
      WasPrivate := False;
      WasPersonal := False;
      PrivNames := '';
      PersNames := '';
      PrivMsgs := '';
      PersMsgs := '';
      //   ...
      OL := LoadAndParseText(ToSay);

      // ...  
      if (OL = nil) then
      begin
        // .  
        SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_CHAT_FAILED, 0);
      end
      else
      begin
        //    
        SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_CHAT_OK, Integer(OL));

        // ,    ,    ...

        //    
        while (PrivNames <> '') do
        begin
          //    ...
          Name := GetNextLine(PrivNames);
          Msg := GetNextLine(PrivMsgs);
          //   
          SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_PRIVATE, StrToBuf(Name + #10 + Msg));
        end;

        //    
        while (PersNames <> '') do
        begin
          //    ...
          Name := GetNextLine(PersNames);
          Msg := GetNextLine(PersMsgs);
          //   
          SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_PERSONAL, StrToBuf(Name + #10 + Msg));
        end;

        //     
        while (ArrivedNames <> '') do
        begin
          //  
          SendNotifyMessage(SockU.FormToNotify, WM_TAKEVENT, TAKEVENT_ARRIVES,
            StrToBuf(GetNextLine(ArrivedNames)));
        end;
      end;
    end;

    if Terminated then break;  //   ?

    //   ,      ,  ,
    //      
    Sleep(100);
  until Terminated;
except on
  E: Exception do
    MessageBox(0, PChar(E.Message), 'ERROR!', MB_OK);
end;
end;

procedure TRefreshThread.Lock;
begin
  EnterCriticalSection(CS);
end;

procedure TRefreshThread.Unlock;
begin
  LeaveCriticalSection(CS);
end;

procedure TRefreshThread.Say(const Msg: string);
begin
  if (Msg = '') then exit; //    
  Lock;   //  
  FSayList.Append(Msg);  //  
  Unlock; //    
end;

procedure TRefreshThread.ForceOnlineRefresh;
begin
  //   ...
  InterlockedExchange(FRefreshOnline, 1);
end;

procedure TRefreshThread.ForceChatRefresh;
begin
  // ... API (,  , ?)
  InterlockedExchange(FRefreshChat, 1);
end;

procedure TRefreshThread.SetNewIntervals(ChatUpdateInterval,
  OnlineUpdateInterval: Integer);
begin
  Lock;   //  
  FChatUpdateInterval := DWORD(ChatUpdateInterval) * 1000;
  FOnlineUpdateInterval := DWORD(OnlineUpdateInterval) * 60 * 1000;
  Unlock; //    
end;

end.
