{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y-,Z1}
{.$OPTIMIZATION OFF}
unit TAKShowU;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics, StdCtrls, Forms;


type
  TReaderProsition = record
    Para, Glyph: Integer;
  end;

  TGlyphFactory = class;

{$HINTS OFF}
  TRefObject = class
  protected
    FRefCount: Integer;
    FFactory: TGlyphFactory;

    constructor Create;
    destructor Destroy; override;
    function Clone: TRefObject;

  public
    procedure Release;
  end;
{$HINTS ON}


  TGlyphStyleRec = record
    FontStyle: TFontStyles;
    FontFace: string[128];
    FontColor: TColor;
    FontSize: Integer;
  end;

  TGlyphStyle = class(TRefObject)
  private
    FStyle: TGlyphStyleRec;
    FFont: TFont;

  protected
    constructor Create(const pStyle: TGlyphStyleRec);
    destructor Destroy; override;

  public
    function IsEqual(const pStyle: TGlyphStyleRec): Boolean;

    property Style: TGlyphStyleRec read FStyle;
  end;


  TGlyph = class;
  TGlyphClickNotify = procedure(Sender: TGlyph; ReaderPos: TReaderProsition) of object;
  TGlyph = class(TRefObject)
  private
    FStyle: TGlyphStyle;
    FWidth: Integer;
    FHeight: Integer;

    FOnClick: TGlyphClickNotify;
    FCursor: TCursor;

    FUserData: Integer;

  protected
    constructor Create(const pStyle: TGlyphStyle; pOnClick: TGlyphClickNotify;
      pUserData: Integer; pCursor: TCursor);
    destructor Destroy; override;

    procedure Draw(X, Y: Integer; Canvas: TCanvas); virtual;

  public
    property Cursor: TCursor read FCursor write FCursor;
    property Style: TGlyphStyle read FStyle;
    property Width: Integer read FWidth;
    property Height: Integer read FHeight;
    property UserData: Integer read FUserData;

    //events
    property OnClick: TGlyphClickNotify read FOnClick;
  end;

  TGlyphChar = class(TGlyph)
  private
    FChar: Char;

  protected
    constructor Create(const pCh: Char; const pStyle: TGlyphStyle;
      pOnClick: TGlyphClickNotify; pUserData: Integer; pCursor: TCursor);

    procedure Draw(X, Y: Integer; Canvas: TCanvas); override;
  end;

  TGlyphBmp = class(TGlyph)
  private
    FBmp: TBitmap;
    FName: string;

  protected
    constructor Create(const Name: string);
    destructor Destroy; override;

    procedure Draw(X, Y: Integer; Canvas: TCanvas); override;
  end;


  TGlyphFactory = class
  protected
    FObjList: TList;

    procedure Clear;
    procedure Add(Obj: TRefObject);
    procedure Delete(Obj: TRefObject);
    function Find(Obj: TRefObject): Integer;

  public
    constructor Create;
    destructor Destroy; override;
  end;

  TGlyphStyleFactory = class(TGlyphFactory)
  public
    function ProduceStyle(const Style: TGlyphStyleRec): TGlyphStyle;
  end;

  TGlyphCharFactory = class(TGlyphFactory)
  public
    function ProduceChar(const Ch: Char; const Style: TGlyphStyleRec;
      pOnClick: TGlyphClickNotify; pUserData: Integer; pCursor: TCursor): TGlyphChar;
  end;

  TGlyphBmpFactory = class(TGlyphFactory)
  public
    function ProduceBmp(const Name: string): TGlyphBmp;
  end;


  TTAKParagraph = class
  private
    FGlyphList: TList;

    procedure Clear;

  public
    constructor Create;
    destructor Destroy; override;

    procedure AddGlyph(Glyph: TGlyph);  // doesn't clone glyph

    function GetGlyph(Index: Integer): TGlyph;
    function GetGlyphCount: Integer;
  end;

  TTAKLine = class
  private
    FYOfs: Integer;
    FHeight: Integer;
    FWidth: Integer;
    FGlyphList: TList;
    FAfter: Integer;

    FParNum: Integer;
    FFirstGlyphNum: Integer;

    FPara: TTAKParagraph;

    procedure Clear;

  public
    constructor Create(pPara: TTAKParagraph; PNum, GNum: Integer);
    destructor Destroy; override;

    procedure AddGlyph(Glyph: TGlyph);  // clones glyph

    procedure Draw(X, Y, Width: Integer; Canvas: TCanvas); virtual;
  end;

  TTAKDocument = class;
  TTAKReader = class
  private
    FText: TTAKDocument;
    FParNum, FGlyphNum: Integer;

    function GetCurGlyph: TGlyph;

{$HINTS OFF}
    constructor Create(pText: TTAKDocument);
{$HINTS ON}

  public
    destructor Destroy; override;

    procedure SetPosition(NewPosition: TReaderProsition);
    function GetPosition: TReaderProsition;

    function PeekChar: Char; // #10 -- EOP, #31 -- special, #27 -- eof
    function GetChar: Char;

    procedure SkipChar;
    procedure PrevChar;

    function IsUnderlined: Boolean;
    function IsBold: Boolean;
    function IsItalic: Boolean;
    function IsStrikeout: Boolean;
    function IsSpaceOrSpecial: Boolean;
    function GetFontFace: string;
    function GetFontSize: Integer;
    function GetFontColor: TColor;

    function GetUnderlinedText: string;

    function EOF: Boolean;
    function EOP: Boolean;
    function BOF: Boolean;
  end;

  TTAKDocument = class
  private
    FParas: TList;
    FLines: TList;

    FWidth: Integer;

    FInterParas: Integer;
    FInterLines: Integer;

    procedure ClearParas;
    procedure ClearLines;

    procedure FormatParagraph(Para: TTAKParagraph; PNum: Integer);

  public
    constructor Create(pInterParas, pInterLines: Integer);
    destructor Destroy; override;

    procedure AddParagraph(Para: TTAKParagraph);
    procedure Reformat(NewWidth: Integer; Forced: Boolean);

    procedure Draw(X, Y: Integer; TopLine, Width, Height: Integer; Canvas: TCanvas);
    function GetGlyphAt(X, Y: Integer; TopLine, Width, Height: Integer; out ReaderPos: TReaderProsition): TGlyph;

    function ParagraphCount: Integer;
    procedure DeleteParagraph(Index: Integer);

    function NewReader: TTAKReader;
  end;

  TTAKView = class(TCustomControl)
  private
    FText: TTAKDocument;
    FStyle: TGlyphStyleRec;
    FPara: TTAKParagraph;

    FSBar: TScrollBar;

    procedure WMSize(var Message: TMessage); message WM_SIZE;
    procedure WMMMove(var Message: TMessage); message WM_MOUSEMOVE;
    procedure WMLBUp(var Message: TMessage); message WM_LBUTTONUP;
    procedure ScrollBarChange(Sender: TObject);

    procedure AdjSBar;

  protected
    procedure Paint; override;

    function LastVisLines: Integer;

    procedure CreateParams(var Params: TCreateParams); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ToBottom;
    procedure Reformat;

    procedure OpenParagraph;
    procedure CloseParagraph;
    procedure SetStyle(const Style: TGlyphStyleRec);
    procedure AddString(const S: string);
    procedure AddImage(const Name: string);
    procedure AddActionString(const S: string; Action: TGlyphClickNotify; Data: Integer);
    procedure AddActionStringWithStyle(const S: string; Style: TGlyphStyleRec; Action: TGlyphClickNotify; Data: Integer);

    property Text: TTAKDocument read FText;
  end;


function GlyphStyleRec(const pFontStyle: TFontStyles; const pFontColor: TColor;
  const pFontFace: string; const pFontSize: Integer): TGlyphStyleRec;


var
  GlyphStyleFactory: TGlyphStyleFactory = nil;
  GlyphCharFactory: TGlyphCharFactory = nil;
  GlyphBmpFactory: TGlyphBmpFactory = nil;


implementation


procedure Register;
begin
  RegisterComponents('Samples', [TTAKView]);
end;

function GlyphStyleRec(const pFontStyle: TFontStyles; const pFontColor: TColor;
  const pFontFace: string; const pFontSize: Integer): TGlyphStyleRec;
begin
  Result.FontStyle := pFontStyle;
  Result.FontFace := pFontFace;
  Result.FontColor := pFontColor;
  Result.FontSize := pFontSize;
end;


{ TRefObject }
constructor TRefObject.Create;
begin
  FFactory := nil;
  FRefCount := 1;
  inherited;
end;

destructor TRefObject.Destroy;
begin
  inherited;
end;

procedure TRefObject.Release;
begin
  Assert(FRefCount > 0, 'TRefObj: invalid RefCount');
  Dec(FRefCount);
  if (FRefCount = 0) then
  begin
    if (FFactory <> nil) then
      FFactory.Delete(Self)
    else
      Destroy;
  end;
end;

function TRefObject.Clone: TRefObject;
begin
  Inc(FRefCount);
  Result := Self;
end;


{ TGlyphStyle }
constructor TGlyphStyle.Create(const pStyle: TGlyphStyleRec);
begin
  inherited Create;

  FStyle := pStyle;

  FFont := TFont.Create;
  FFont.Name := FStyle.FontFace;
  FFont.Size := FStyle.FontSize;
  FFont.Color := FStyle.FontColor;
  FFont.Style := FStyle.FontStyle;
  FFont.CharSet := RUSSIAN_CHARSET;
end;

destructor TGlyphStyle.Destroy;
begin
  FFont.Free;
  inherited;
end;

function TGlyphStyle.IsEqual(const pStyle: TGlyphStyleRec): Boolean;
begin
  Result := (pStyle.FontFace = FStyle.FontFace) and
    (pStyle.FontSize = FStyle.FontSize) and
    (pStyle.FontColor = FStyle.FontColor) and
    (pStyle.FontStyle = FStyle.FontStyle);
end;


{ TGlyph }
constructor TGlyph.Create(const pStyle: TGlyphStyle; pOnClick: TGlyphClickNotify;
  pUserData: Integer; pCursor: TCursor);
begin
  Assert(pStyle <> nil, 'TGlyph.Create: pStyle is nil');
  FRefCount := 1;

  FWidth := 0;
  FHeight := 0;

  FOnClick := pOnCLick;
  FCursor := pCursor;
  FStyle := pStyle;
  FUserData := pUserData;

  inherited Create;
end;

destructor TGlyph.Destroy;
begin
  FStyle.Release;
  inherited;
end;

procedure TGlyph.Draw(X, Y: Integer; Canvas: TCanvas);
begin
  Assert(Canvas <> nil, 'TGlyph.Draw: Canvas is nil');
end;


{ TGlyphChar }
constructor TGlyphChar.Create(const pCh: Char; const pStyle: TGlyphStyle;
  pOnClick: TGlyphClickNotify; pUserData: Integer; pCursor: TCursor);
var
  B: TBitmap;
  Sz: TSize;
begin
  inherited Create(pStyle, pOnClick, pUserData, pCursor);
  FChar := pCh;

  B := TBitmap.Create;
  B.Canvas.Font := pStyle.FFont;
  Sz := B.Canvas.TextExtent(FChar);
  B.Free;

  FWidth := Sz.cx + 1;
  FHeight := Sz.cy;
end;

procedure TGlyphChar.Draw(X, Y: Integer; Canvas: TCanvas);
begin
  Canvas.Font := FStyle.FFont;
  Canvas.TextOut(X, Y, FChar);

  //    --    
  //Canvas.FillRect(Rect(X + FWidth - 1, Y, X + FWidth, Y + FHeight));
end;


{ TGlyphBmp }
constructor TGlyphBmp.Create(const Name: string);
var
  Sz: TSize;
begin
  inherited Create(GlyphStyleFactory.ProduceStyle(GlyphStyleRec([], clBlue, 'Tahoma', 8)), nil, 0,
    crDefault);
  FName := Name;
  FBmp := TBitmap.Create;
  try
    FBmp.LoadFromFile('bmp\' + FName + '.bmp');
    FBmp.TransparentColor
  except
    FBmp.Canvas.Font := FStyle.FFont;
    Sz := FBmp.Canvas.TextExtent(':' + FName + ':');
    FBmp.Width := Sz.cx;
    FBmp.Height := Sz.cy;
    FBmp.Canvas.TextOut(0, 0, ':' + FName + ':');
  end;
  FWidth := FBmp.Width;
  FHeight := FBmp.Height;
end;

destructor TGlyphBmp.Destroy;
begin
  FreeAndNil(FBmp);
  inherited Destroy;
end;

procedure TGlyphBmp.Draw(X, Y: Integer; Canvas: TCanvas);
begin
  Canvas.Draw(X, Y, FBmp);
end;


{ TGlyphFactory }
constructor TGlyphFactory.Create;
begin
  inherited;
  FObjList := TList.Create;
end;

destructor TGlyphFactory.Destroy;
begin
  Clear;
  FObjList.Free;
  inherited;
end;

procedure TGlyphFactory.Clear;
var
  F: Integer;
begin
  for F := 0 to FObjList.Count - 1 do
    if (FObjList[F] <> nil) then
    begin
      TRefObject(FObjList[F]).Free;
      FObjList[F] := nil;
    end;
  FObjList.Clear;
end;

function TGlyphFactory.Find(Obj: TRefObject): Integer;
begin
  Result := 0;
  while (Result < FObjList.Count) and (FObjList[Result] <> Obj) do Inc(Result);
  if (Result = FObjList.Count) then Result := -1;
end;

procedure TGlyphFactory.Add(Obj: TRefObject);
var
  F: Integer;
begin
  F := Find(Obj);
  Assert(F = -1, 'TGlyphFactory: object duplication');
  F := Find(nil);
  if (F = -1) then FObjList.Add(Obj) else FObjList[F] := Obj;
  Obj.FFactory := Self;
end;

procedure TGlyphFactory.Delete(Obj: TRefObject);
var
  F: Integer;
begin
  F := Find(Obj);
  Assert(F <> -1, 'TGlyphFactory: deleting inexistent object');
  TRefObject(FObjList[F]).Free;
  FObjList[F] := nil;
end;


{ TGlyphStyleFactory }
function TGlyphStyleFactory.ProduceStyle(const Style: TGlyphStyleRec): TGlyphStyle;
var
  F: Integer;
begin
  for F := 0 to FObjList.Count - 1 do
  begin
    if (FObjList[F] <> nil) and
       TGlyphStyle(FObjList[F]).IsEqual(Style) then
    begin
      Inc(TRefObject(FObjList[F]).FRefCount);
      Result := TGlyphStyle(FObjList[F]);
      exit;
    end;
  end;

  Result := TGlyphStyle.Create(Style);
  Add(Result);
end;


{ TGlyphCharFactory }
function TGlyphCharFactory.ProduceChar(const Ch: Char; const Style: TGlyphStyleRec;
  pOnClick: TGlyphClickNotify; pUserData: Integer; pCursor: TCursor): TGlyphChar;
var
  F: Integer;
  StyleO: TGlyphStyle;
begin
  for F := 0 to FObjList.Count - 1 do
  begin
    if (FObjList[F] <> nil) and
       (TGlyphChar(FObjList[F]).FChar = Ch) and
       (@TGlyphChar(FObjList[F]).FOnClick = @pOnClick) and
       (TGlyphChar(FObjList[F]).FUserData = pUserData) and
       (TGlyphChar(FObjList[F]).FCursor = pCursor) and
       TGlyphChar(FObjList[F]).FStyle.IsEqual(Style) then
    begin
      Inc(TRefObject(FObjList[F]).FRefCount);
      Result := TGlyphChar(FObjList[F]);
      exit;
    end;
  end;

  StyleO := GlyphStyleFactory.ProduceStyle(Style);
  Result := TGlyphChar.Create(Ch, StyleO, pOnClick, pUserData, pCursor);
  Add(Result);
end;


{ TGlyphBmpFactory }
function TGlyphBmpFactory.ProduceBmp(const Name: string): TGlyphBmp;
var
  F: Integer;
begin
  for F := 0 to FObjList.Count - 1 do
  begin
    if (FObjList[F] <> nil) and
       (TGlyphBmp(FObjList[F]).FName = Name) then
    begin
      Inc(TRefObject(FObjList[F]).FRefCount);
      Result := TGlyphBmp(FObjList[F]);
      exit;
    end;
  end;

  Result := TGlyphBmp.Create(Name);
  Add(Result);
end;


{ TTAKParagraph }
constructor TTAKParagraph.Create;
begin
  inherited Create;
  FGlyphList := TList.Create;
end;

destructor TTAKParagraph.Destroy;
begin
  Clear;
  FGlyphList.Free;
  inherited Destroy;
end;

procedure TTAKParagraph.Clear;
var
  F: Integer;
begin
  for F := 0 to FGlyphList.Count - 1 do TGlyph(FGlyphList[F]).Release;
  FGlyphList.Clear;
end;

procedure TTAKParagraph.AddGlyph(Glyph: TGlyph);
begin
  Assert(Glyph <> nil, 'TTAKParagraph.AddGlyph: Glyph is nil');
  FGlyphList.Add(Glyph);
end;

function TTAKParagraph.GetGlyph(Index: Integer): TGlyph;
begin
  Assert((Index >= 0) and (Index < FGlyphList.Count), 'TTAKParagraph.GetGlyph: invalid index');
  Result := TGlyph(FGlyphList[Index]);
end;

function TTAKParagraph.GetGlyphCount: Integer;
begin
  Result := FGlyphList.Count;
end;


{ TTAKLine }
constructor TTAKLine.Create(pPara: TTAKParagraph; PNum, GNum: Integer);
begin
  Assert(pPara <> nil, 'TTAKLine.Create: pPara is nil');
  inherited Create;
  FPara := pPara;
  FYOfs := 0;
  FHeight := 0;
  FWidth := 0;
  FGlyphList := TList.Create;
  FParNum := PNum;
  FFirstGlyphNum := GNum;
end;

destructor TTAKLine.Destroy;
begin
  Clear;
  FGlyphList.Free;
  inherited Destroy;
end;

procedure TTAKLine.Clear;
var
  F: Integer;
begin
  for F := 0 to FGlyphList.Count - 1 do TGlyph(FGlyphList[F]).Release;
  FGlyphList.Clear;
end;

procedure TTAKLine.AddGlyph(Glyph: TGlyph);
var
  H: Integer;
begin
  Assert(Glyph <> nil, 'TTAKLine.AddGlyph: Glyph is nil');
  FGlyphList.Add(Glyph.Clone);
  H := Glyph.FHeight;
  if (H > FHeight) then FHeight := H;
  Inc(FWidth, Glyph.FWidth);
end;

procedure TTAKLine.Draw(X, Y, Width: Integer; Canvas: TCanvas);
var
  F: Integer;
  O: TGlyph;
begin
  Assert(Canvas <> nil, 'TTAKLine.Draw: Canvas is nil');
  F := 0;
  while (Width > 0) and (F < FGlyphList.Count) do
  begin
    O := TGlyph(FGlyphList[F]);
    O.Draw(X, Y + FHeight - O.FHeight, Canvas);
    Inc(X, O.FWidth);
    Dec(Width, O.FWidth);
    Inc(F);
  end;
end;


{ TTAKDocument }
constructor TTAKDocument.Create(pInterParas, pInterLines: Integer);
begin
  inherited Create;
  FParas := TList.Create;
  FLines := TList.Create;

  FInterParas := pInterParas;
  FInterLines := pInterLines;

  FWidth := 0;
end;

destructor TTAKDocument.Destroy;
begin
  ClearParas;
  ClearLines;
  FParas.Free;
  FLines.Free;
  inherited Destroy;
end;

procedure TTAKDocument.ClearParas;
var
  F: Integer;
begin
  for F := 0 to FParas.Count - 1 do TTAKParagraph(FParas[F]).Free;
  FParas.Clear;
end;

procedure TTAKDocument.ClearLines;
var
  F: Integer;
begin
  for F := 0 to FLines.Count - 1 do TTAKLine(FLines[F]).Free;
  FLines.Clear;
end;

procedure TTAKDocument.AddParagraph(Para: TTAKParagraph);
begin
  Assert(Para <> nil, 'TTAKDocument.AddParagraph: Para is nil');
  FParas.Add(Para);
  FormatParagraph(Para, FParas.Count - 1);
end;

procedure TTAKDocument.Reformat(NewWidth: Integer; Forced: Boolean);
var
  F: Integer;
begin
  if not Forced and (FWidth = NewWidth) then exit;
  ClearLines;
  FWidth := NewWidth;
  for F := 0 to FParas.Count - 1 do FormatParagraph(TTAKParagraph(FParas[F]), F);
end;

function IsSpaceOrNotChar(Glyph: TGlyph): Boolean;
begin
  Assert(Glyph <> nil, 'IsSpaceOrNotChar: Glyph = nil');
  if (Glyph is TGlyphChar) then
  begin
    Result := ((Glyph as TGlyphChar).FChar = ' ');
  end
  else
    Result := True;
end;

function IsSpace(Glyph: TGlyph): Boolean;
begin
  Assert(Glyph <> nil, 'IsSpace: Glyph = nil');
  if (Glyph is TGlyphChar) then
  begin
    Result := ((Glyph as TGlyphChar).FChar = ' ');
  end
  else
    Result := False;
end;

procedure TTAKDocument.FormatParagraph(Para: TTAKParagraph; PNum: Integer);
var
  YOfs: Integer;
  L: TTAKLine;
  PrevWdt, WStart, F, E, W: Integer;
begin
  Assert(Para <> nil, 'TTAKDocument.FormatParagraph: Para is nil');
  if (FWidth < 1) then exit;

  L := nil;
  if (FLines.Count = 0) then
    YOfs := 0
  else
  begin
    L := TTAKLine(FLines[FLines.Count - 1]);
    L.FAfter := FInterParas;
    YOfs := L.FYOfs + L.FHeight + L.FAfter;
  end;

  F := 0;
  while (F < Para.GetGlyphCount) do
  begin
    E := F;
    W := 0;

    // count glyphs that will fill the line
    while (E < Para.GetGlyphCount) do
    begin
      WStart := E;
      PrevWdt := W;
      if IsSpaceOrNotChar(Para.GetGlyph(E)) then
      begin
        //         --   
        Inc(W, Para.GetGlyph(E).Width);
        Inc(E);
      end
      else
      begin
        //  . .       
        //    .   
        while (E < Para.GetGlyphCount) and not IsSpaceOrNotChar(Para.GetGlyph(E)) do
        begin
          Inc(W, Para.GetGlyph(E).Width);
          Inc(E);
        end;
      end;

      //    E -- ,    , WStart --   
      if (W > FWidth) then
      begin
        //  -  !
        if (WStart = F) then
        begin
          //      ...
          if (E - WStart = 1) then
          begin
            //    ( , !),
            //     --     .  .  .
            break;
          end
          else
          begin
            //  --   ,  
            //    -- ,      
            E := WStart;
            W := PrevWdt;
            while (E < Para.GetGlyphCount) and (W <= FWidth) do
            begin
              Inc(W, Para.GetGlyph(E).Width);
              Inc(E);
            end;
            if (E <> WStart + 1) then Dec(E);
            break;
          end;
        end
        else
        begin
          E := WStart;
          break;  //      ?
        end;
      end;
    end;

    //   
    L := TTAKLine.Create(Para, PNum, F);
    L.FYOfs := YOfs;
    while (F < E) do
    begin
      L.AddGlyph(Para.GetGlyph(F));
      Inc(F);
    end;
    L.FAfter := FInterLines;
    FLines.Add(L);
    Inc(YOfs, L.FHeight + L.FAfter);
  end;

  if (L <> nil) then L.FAfter := 0;
end;

procedure TTAKDocument.Draw(X, Y: Integer; TopLine, Width, Height: Integer; Canvas: TCanvas);
var
  R: TRect;
  L: TTAKLine;
  H: Integer;
begin
  Assert(Canvas <> nil, 'TTAKDocument.Draw: Canvas is nil');
  Canvas.Brush.Style := bsSolid;
  Canvas.Brush.Color := $00EFEFF7;
  Canvas.FillRect(Rect(X, Y, X + Width, Y + Height));

  H := 0;
  while (H < Height) and (TopLine < FLines.Count) do
  begin
    L := TTAKLine(FLines[TopLine]);
    L.Draw(X, Y, Width, Canvas);

    Inc(H, L.FHeight + L.FAfter);
    Inc(Y, L.FHeight + L.FAfter);
    Inc(TopLine);
  end;

  R.Top := H;
  R.Left := 0;
  R.Bottom := Height;
  R.Right := Width;
  Canvas.FillRect(R);
end;

function TTAKDocument.GetGlyphAt(X, Y, TopLine, Width,
  Height: Integer; out ReaderPos: TReaderProsition): TGlyph;
var
  L: TTAKLine;
  F, H, P: Integer;
begin
  Result := nil;
  H := 0;
  ReaderPos.Para := 0;
  ReaderPos.Glyph := 0;

  while (H < Height) and (TopLine < FLines.Count) do
  begin
    L := TTAKLine(FLines[TopLine]);
    if (Y >= H) and (Y < H + L.FHeight + L.FAfter) then
    begin
      P := 0;
      for F := 0 to L.FGlyphList.Count - 1 do
      begin
        if (X >= P) and (X <= P + TGlyph(L.FGlyphList[F]).FWidth) then
        begin
          Result := TGlyph(L.FGlyphList[F]);
          ReaderPos.Para := L.FParNum;
          ReaderPos.Glyph := L.FFirstGlyphNum + F;
          break;
        end;
        Inc(P, TGlyph(L.FGlyphList[F]).FWidth);
      end;
      exit;
    end;

    Inc(H, L.FHeight + L.FAfter);
    Inc(TopLine);
  end;
end;

function TTAKDocument.NewReader: TTAKReader;
begin
  Result := TTAKReader.Create(Self);
end;

procedure TTAKDocument.DeleteParagraph(Index: Integer);
begin
  Assert((Index >= 0) and (Index <= FParas.Count), 'TTAKDocument.DeleteParagraph: invalid index');
  TTAKParagraph(FParas[Index]).Free;
  FParas.Delete(Index);
end;

function TTAKDocument.ParagraphCount: Integer;
begin
  Result := FParas.Count;
end;


{ TTAKView }
constructor TTAKView.Create(AOwner: TComponent);
begin
  inherited;
  DoubleBuffered := True;

  FText := TTAKDocument.Create(2, 1);
  FPara := nil;
  FStyle.FontStyle := [];
  FStyle.FontFace := 'Tahoma';
  FStyle.FontColor := clWindowText;
  FStyle.FontSize := 10;

  FSBar := TScrollBar.Create(Self);
  FSBar.Parent := Self;
  FSBar.Kind := sbVertical;
  FSBar.Align := alRight;
  FSBar.Enabled := False;
  FSBar.OnChange := ScrollBarChange;

  Reformat;
end;

destructor TTAKView.Destroy;
begin
  FreeAndNil(FPara);
  FreeAndNil(FText);
  inherited;
end;

procedure TTAKView.AddString(const S: string);
begin
  AddActionString(S, nil, 0);
end;

procedure TTAKView.AddActionString(const S: string; Action: TGlyphClickNotify; Data: Integer);
var
  G: TGlyph;
  F: Integer;
  Cr: TCursor;
begin
  Assert(FPara <> nil, 'TTAKView.AddActionString: paragraph was not opened');
  if Assigned(Action) then Cr := crCross else Cr := crDefault;
  for F := 1 to Length(S) do
  begin
    G := GlyphCharFactory.ProduceChar(S[F], FStyle, Action, Data, Cr);
    FPara.AddGlyph(G);
  end;
end;

procedure TTAKView.AddActionStringWithStyle(const S: string; Style: TGlyphStyleRec; Action: TGlyphClickNotify; Data: Integer);
var
  G: TGlyph;
  F: Integer;
  Cr: TCursor;
begin
  Assert(FPara <> nil, 'TTAKView.AddActionString: paragraph was not opened');
  if (S = '') then exit;
  if Assigned(Action) then Cr := crCross else Cr := crDefault;
  for F := 1 to Length(S) do
  begin
    G := GlyphCharFactory.ProduceChar(S[F], Style, Action, Data, Cr);
    FPara.AddGlyph(G);
  end;
end;

procedure TTAKView.AddImage(const Name: string);
var
  G: TGlyph;
begin
  Assert(FPara <> nil, 'TTAKView.AddImage: paragraph was not opened');
  G := GlyphBmpFactory.ProduceBmp(Name);
  FPara.AddGlyph(G);
end;

procedure TTAKView.CloseParagraph;
begin
  Assert(FText <> nil, 'TTAKView.CloseParagraph: no text object');
  Assert(FPara <> nil, 'TTAKView.CloseParagraph: paragraph was not opened');
  FText.AddParagraph(FPara);
  FPara := nil;
  ToBottom;
end;

procedure TTAKView.OpenParagraph;
begin
  Assert(FText <> nil, 'TTAKView.OpenParagraph: no text object');
  FPara := TTAKParagraph.Create;
end;

procedure TTAKView.SetStyle(const Style: TGlyphStyleRec);
begin
  FStyle := Style;
end;

procedure TTAKView.WMSize(var Message: TMessage);
begin
  inherited;
  Reformat;
end;

procedure TTAKView.WMMMove(var Message: TMessage);
var
  G: TGlyph;
  X, Y: Integer;
  RP: TReaderProsition;
begin
  X := (Message.lParam and $FFFF);
  Y := ((Message.lParam shr 16) and $FFFF);
  G := FText.GetGlyphAt(X, Y, FSBar.Position, Width - FSBar.Width, Height, RP);
  if (G = nil) then
    Cursor := crDefault
  else
    Cursor := G.Cursor;
  inherited;
end;

procedure TTAKView.WMLBUp(var Message: TMessage);
var
  G: TGlyph;
  X, Y: Integer;
  RP: TReaderProsition;
begin
  X := (Message.lParam and $FFFF);
  Y := ((Message.lParam shr 16) and $FFFF);
  G := FText.GetGlyphAt(X, Y, FSBar.Position, Width - FSBar.Width, Height, RP);
  if (G <> nil) and Assigned(G.FOnClick) then G.FOnClick(G, RP);
  inherited;
end;

procedure TTAKView.Paint;
begin
  inherited Paint;
  if (FText <> nil) then
    FText.Draw(0, 0, FSBar.Position, Width - FSBar.Width, Height, Canvas);
end;

function TTAKView.LastVisLines: Integer;
var
  L: TTAKLine;
  F, H: Integer;
begin
  H := Height;
  F := FText.FLines.Count - 1;
  while (H >= 0) and (F >= 0) do
  begin
    L := TTAKLine(FText.FLines[F]);
    Dec(H, L.FHeight + L.FAfter);
//    if (H > 0) then Dec(H, L.FAfter);
    if (H >= 0) then Dec(F);
  end;
  Inc(F);

  Result := F;
end;

procedure TTAKView.ScrollBarChange(Sender: TObject);
begin
  Invalidate;
end;

procedure TTAKView.AdjSBar;
var
  LL: Integer;
begin
  LL := LastVisLines;
  FSBar.Max := LL;
  FSBar.Position := LL;
  FSBar.PageSize := 10;
  FSBar.Enabled := (LL > 0);
end;

procedure TTAKView.ToBottom;
begin
  AdjSBar;
  Invalidate;
end;

procedure TTAKView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or WS_CLIPCHILDREN or WS_CLIPSIBLINGS;
end;

procedure TTAKView.Reformat;
begin
  if (FText <> nil) then FText.Reformat(Width - FSBar.Width, True);
  ToBottom;
end;


{ TTAKReader }
constructor TTAKReader.Create(pText: TTAKDocument);
begin
  inherited Create;
  Assert(pText <> nil, 'TTAKReader.Create: pText is nil');
  FText := pText;
  FParNum := 0;
  FGlyphNum := 0;
end;

destructor TTAKReader.Destroy;
begin
  FText := nil;
  inherited;
end;

function TTAKReader.EOF: Boolean;
begin
  Result := (FParNum >= FText.FParas.Count);
end;

function TTAKReader.BOF: Boolean;
begin
  Result := (FparNum = 0) and (FGlyphNum = 0);
end;

function TTAKReader.EOP: Boolean;
begin
  Result := (FParNum >= FText.FParas.Count)
    or (FGlyphNum >= TTAKParagraph(FText.FParas[FParNum]).GetGlyphCount);
end;

function TTAKReader.GetCurGlyph: TGlyph;
begin
  if EOP then
    Result := nil
  else
  begin
    Result := TTAKParagraph(FText.FParas[FParNum]).GetGlyph(FGlyphNum);
{    AllocConsole;
    Write('Pos: ', FParNum, ', ', FGlyphNum, '  ');
    if (Result is TGlyphChar) then
      WriteLn('Char: "', (Result as TGlyphChar).FChar, '"')
    else
      WriteLn('non-char');}
  end;
end;

function TTAKReader.PeekChar: Char;
var
  G: TGlyph;
begin
  if EOF then Result := #27
  else if EOP then Result := #10
  else
  begin
    G := GetCurGlyph;
    Assert(G <> nil, 'TTAKReader.PeekChar: current glyph is nil');
    if (G is TGlyphChar) then
    begin
      Result := (G as TGlyphChar).FChar;
      if (Result in [#0, #10, #27, #31]) then Result := #1;
    end
    else
      Result := #31;
  end;
end;

procedure TTAKReader.SkipChar;
begin
  if EOF then exit;
  if EOP then
  begin
    Inc(FParNum);
    FGlyphNum := 0;
  end
  else
    Inc(FGlyphNum);
end;

procedure TTAKReader.PrevChar;
begin
  if BOF then exit;
  if (FGlyphNum = 0) then
  begin
    Dec(FParNum);
    FGlyphNum := TTAKParagraph(FText.FParas[FParNum]).GetGlyphCount + 1;
  end
  else
    Dec(FGlyphNum);
end;

function TTAKReader.GetChar: Char;
begin
  Result := PeekChar;
  SkipChar;
end;

procedure TTAKReader.SetPosition(NewPosition: TReaderProsition);
begin
  // each paragraph has GetClyphCount + 1 glyph (the last is inexistent EOP one)
  Assert(NewPosition.Para >= 0, 'TTAKReader.SetPosition: NewPostion.Para is negative');
  Assert(NewPosition.Glyph >= 0, 'TTAKReader.SetPosition: NewPostion.Glyph is negative');
  if (NewPosition.Para >= FText.FParas.Count) then
  begin
    NewPosition.Para := FText.FParas.Count;
    NewPosition.Glyph := 0;
  end
  else
  if (NewPosition.Glyph >= TTAKParagraph(FText.FParas[NewPosition.Para]).GetGlyphCount) then
    NewPosition.Glyph := TTAKParagraph(FText.FParas[NewPosition.Para]).GetGlyphCount;
  FParNum := NewPosition.Para;
  FGlyphNum := NewPosition.Glyph;
end;

function TTAKReader.GetPosition: TReaderProsition;
begin
  Result.Para := FParNum;
end;

function TTAKReader.GetFontColor: TColor;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := 0
  else
    Result := G.FStyle.FStyle.FontColor;
end;

function TTAKReader.GetFontFace: string;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := ''
  else
    Result := G.FStyle.FStyle.FontFace;
end;

function TTAKReader.GetFontSize: Integer;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := 0
  else
    Result := G.FStyle.FStyle.FontSize;
end;

function TTAKReader.IsBold: Boolean;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := False
  else
    Result := (fsBold in G.FStyle.FStyle.FontStyle);
end;

function TTAKReader.IsItalic: Boolean;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := False
  else
    Result := (fsItalic in G.FStyle.FStyle.FontStyle);
end;

function TTAKReader.IsSpaceOrSpecial: Boolean;
begin
  Result := (PeekChar <= ' ');
end;

function TTAKReader.IsStrikeout: Boolean;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := False
  else
    Result := (fsStrikeout in G.FStyle.FStyle.FontStyle);
end;

function TTAKReader.IsUnderlined: Boolean;
var
  G: TGlyph;
begin
  G := GetCurGlyph;
  if (G = nil) then
    Result := False
  else
    Result := (fsUnderline in G.FStyle.FStyle.FontStyle);
end;

function TTAKReader.GetUnderlinedText: string;
begin
  Result := '';
  while not BOF and IsUnderlined do PrevChar;
  if not IsUnderlined then SkipChar;
  while not EOP and IsUnderlined do
    Result := Result + GetChar;
end;

begin
  GlyphStyleFactory := TGlyphStyleFactory.Create;
  GlyphCharFactory := TGlyphCharFactory.Create;
  GlyphBmpFactory := TGlyphBmpFactory.Create;
end.
