{$A+,B-,C+,D+,E-,F-,G+,H+,I+,J+,K-,L+,M-,N+,O+,P+,Q-,R+,S-,T-,U-,V+,W-,X+,Y-,Z1}
{.$OPTIMIZATION OFF}
unit fmMainU;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ExtCtrls, StdCtrls, AppEvnts, Menus, ShellApi,
  RefreshThreadU, SockU, TAKShowU, ImgList, ActnList;

const
  WM_TRAYMSG = WM_USER + 876;


type
  TOptions = record
    UseProxy: Boolean;
    ProxyAddr: string;
    ProxyPort: Word;
    UseAuth: Boolean;
    AuthUser: string;
    AuthPassword: string;
    LoginName: string;
    LoginPassword: string;
    TrayIcon: Boolean;
    PermanentIcon: Boolean;
    BeepOn: Boolean;
    PopupsOn: Boolean;
    PopupsLeft: Boolean;
    PopupsTop: Boolean;
    DebugLog: Boolean;
    RememberIds: Boolean;
    ChatRefTime: Integer;
    UListRefTime: Integer;
    MasterServer: string;
  end;

  TfmMain = class(TForm)
    StatusBar: TStatusBar;
    Panel2: TPanel;
    AppEvents: TApplicationEvents;
    Splitter: TSplitter;
    ColorDialog: TColorDialog;
    cbText: TComboBox;
    MainMenu: TMainMenu;
    ActionList: TActionList;
    ActionEnter: TAction;
    imIcons: TImageList;
    miChat: TMenuItem;
    miEnter: TMenuItem;
    ActionLeave: TAction;
    miLeave: TMenuItem;
    miN1: TMenuItem;
    ActionSay: TAction;
    miSay: TMenuItem;
    ActionPrefs: TAction;
    miPreferences: TMenuItem;
    miExit: TMenuItem;
    ActionSetColor: TAction;
    miChangeColor: TMenuItem;
    ActionRefresh: TAction;
    miRefresh: TMenuItem;
    btSay: TButton;
    ActionCutLog: TAction;
    miClearLog: TMenuItem;
    lbOnline: TListView;
    procedure lbOnlineDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure AppEventsMinimize(Sender: TObject);
    procedure AppEventsRestore(Sender: TObject);
    procedure RxTrayIconDblClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Panel2Resize(Sender: TObject);
    procedure cbTextKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ActionEnterExecute(Sender: TObject);
    procedure ActionLeaveExecute(Sender: TObject);
    procedure ActionRefreshExecute(Sender: TObject);
    procedure ActionSetColorExecute(Sender: TObject);
    procedure ActionSayExecute(Sender: TObject);
    procedure miExitClick(Sender: TObject);
    procedure ActionPrefsExecute(Sender: TObject);
    procedure ActionCutLogExecute(Sender: TObject);

  private
    TrayIconData: TNotifyIconData;
    TrayIconActive: Boolean;

    Options: TOptions;
    CurColor: TColor;

    View: TTAKView;

    RThread: TRefreshThread;

    procedure SetTrayIconStatus(Active: Boolean);
    procedure SetColor;

    procedure NameClick(Sender: TGlyph; ReaderPos: TReaderProsition);
    procedure MailClick(Sender: TGlyph; ReaderPos: TReaderProsition);
    procedure URLClick(Sender: TGlyph; ReaderPos: TReaderProsition);

    function IsURL(S: string): string;
    function IsMail(S: string): string;

    function FindUserInList(Name: string): Integer;
    procedure AddToHistory(const S: string);

    procedure LogMsg(S: string);
    procedure DisableAll;          //   
    procedure EnableAll;           //   

    // ,    
    procedure MsgStatus(var M: TMessage); message WM_TAKSTATUS;
    procedure MsgEvent(var M: TMessage); message WM_TAKEVENT;
    procedure MsgDeath(var M: TMessage); message WM_TAKTHREADDIES;

    procedure MsgTray(var M: TMessage); message WM_TRAYMSG;

    procedure MsgUser(var M: TMessage); message WM_USER;

    procedure ActivateEdit;

    procedure DoLogin;
    procedure DoLogout;

    procedure SetDefaultOptions;
    function LoadIni: Boolean;
    procedure SaveIni;

    procedure SaveIds;
    procedure LoadIds;
  end;

var
  fmMain: TfmMain;

implementation

uses
  SmilesU, PopupsU, LogU, fmPrefsU;

{$R *.DFM}

{ TfmMain }

procedure DisableAction(C: TAction);
begin
  C.Tag := Byte(C.Enabled);
  C.Enabled := False;
end;

procedure EnableAction(C: TAction);
begin
  C.Enabled := (C.Tag <> 0);
end;

procedure SaveStr(S: TStream; St: string);
var
  L: Integer;
begin
  L := Length(St) xor $55AA55AA;
  S.WriteBuffer(L, 4);
  for L := 1 to Length(St) do St[L] := Char(Byte(St[L]) xor $55);
  if (St <> '') then S.WriteBuffer(St[1], Length(St));
end;

function LoadStr(S: TStream): string;
var
  L: Integer;
begin
  Result := '';
  S.ReadBuffer(L, 4);
  L := L xor $55AA55AA;
  if (L < 0) or (L > 127) then raise Exception.Create('!');
  if (L = 0) then exit;
  SetLength(Result, L);
  S.ReadBuffer(Result[1], L);
  for L := 1 to Length(Result) do Result[L] := Char(Byte(Result[L]) xor $55);
end;

procedure SaveBool(S: TStream; B: Boolean);
var
  Bt: Byte;
begin
  if B then Bt := 255 else Bt := 0;
  S.WriteBuffer(Bt, 1);
end;

function LoadBool(S: TStream): Boolean;
var
  Bt: Byte;
begin
  S.ReadBuffer(Bt, 1);
  Result := (Bt <> 0);
end;

procedure SaveInt(S: TStream; I: Integer);
begin
  S.WriteBuffer(I, SizeOf(I));
end;

function LoadInt(S: TStream): Integer;
begin
  S.ReadBuffer(Result, SizeOf(Result));
end;

procedure TfmMain.MsgDeath(var M: TMessage);
begin
  RThread := nil;
  LogMsg(#31#00#00#255#00#00'thread dies.');
  ActionEnter.Tag := 1;
  ActionLeave.Tag := 0;
  ActionRefresh.Tag := 0;
  ActionSay.Tag := 0;
  EnableAll;
end;

procedure TfmMain.MsgStatus(var M: TMessage);
var
  S: string;
begin
  S := BufToStr(M.LParam);
  StatusBar.SimpleText := S;
  StatusBar.Update;
end;

function GetName(const S: string): string;
begin
  Result := S;
  Delete(Result, 1, Pos(#9, Result));
  Result := Copy(Result, 1, Pos(#0, Result) - 1);
  Delete(Result, 1, 1);
  Delete(Result, Length(Result), 1);
end;

procedure TfmMain.MsgEvent(var M: TMessage);
var
  L: TStringList;
  LI: TListItem;
  St, N: string;
  F, C: Integer;
begin
  case M.WParam of
    TAKEVENT_LOGGING_IN: LogMsg(#31#00#128#0#00#00'logging in...');
    TAKEVENT_LOGIN_FAILED:
    begin
      LogMsg(#31#00#00#255#00#00'login failed.');
      EnableAll;
      ActionEnter.Enabled := True;
    end;
    TAKEVENT_LOGIN_OK:
    begin
      LogMsg(#31#255#00#00#00#00'login ok.');
      ActionEnter.Tag := 0;
      ActionLeave.Tag := 1;
      ActionRefresh.Tag := 1;
      ActionSay.Tag := 1;
      EnableAll;
    end;

    TAKEVENT_LOGGING_OUT: ;
    TAKEVENT_LOGOUT_FAILED: ;
    TAKEVENT_LOGOUT_OK: ;

    TAKEVENT_REFRESHING_ONLINE: ;
    TAKEVENT_ONLINE_FAILED:
      begin
        LogStr('TAKEVENT_ONLINE_FAILED');
        StatusBar.SimpleText := 'TAKEVENT_ONLINE_FAILED';
      end;
    TAKEVENT_ONLINE_OK:
      begin
        L := TStringList(M.LParam);
        lbOnline.Perform(WM_SETREDRAW, 0, 0);
        try
          //   
          F := 0;
          while (F < lbOnline.Items.Count) do
          begin
            C := 0;
            while (C < L.Count) and (ANSICompareText(L[C], lbOnline.Items[F].Caption) <> 0) do Inc(C);
            if (C = L.Count) then
            begin
              //    --  
              if Options.PopupsOn then PopupManager.AddPopup('LEAVES!', lbOnline.Items[F].Caption, POPUP_NORMAL);
              //MessageBeep(DWORD(-1));  //  
              lbOnline.Items.Delete(F)
            end
            else
              Inc(F);  // !
          end;

          //  online-
          for F := 0 to L.Count - 1 do
          begin
            C := FindUserInList(L[F]);
            if (C <> -1) then
              //   ,     (  )
              lbOnline.Items[C].Data := L.Objects[F]
            else
            begin
              // !
              LI := lbOnline.Items.Add;
              LI.Caption := L[F];
              LI.Data := L.Objects[F];
            end;
          end;

        finally
          lbOnline.Perform(WM_SETREDRAW, 1, 0);
          lbOnline.Invalidate;
          L.Free;
        end;
      end;

    TAKEVENT_REFRESHING_CHAT: ;
    TAKEVENT_CHAT_FAILED:
      begin
        LogStr('TAKEVENT_CHAT_FAILED');
        StatusBar.SimpleText := 'TAKEVENT_CHAT_FAILED';
      end;
    TAKEVENT_CHAT_OK:
      begin
        L := TStringList(M.LParam);
        try
          for F := 0 to L.Count - 1 do
          begin
            N := GetName(L[F]);
            C := FindUserInList(N);
            if (C <> -1) and not lbOnline.Items[C].Checked then C := -1;
            if (C = -1) then LogMsg(L[F]);
          end;
        finally
          L.Free;
        end;

        if Options.RememberIds then SaveIds;
      end;

    TAKEVENT_ARRIVES:
      begin
        N := BufToStr(M.LParam);
        if (N <> SockU.ChatName) then
        begin
          if Options.PopupsOn then PopupManager.AddPopup('ARRIVES!', N, POPUP_NORMAL);
        end;
        if Options.BeepOn then MessageBeep(DWORD(-1));
      end;
    TAKEVENT_LEAVES:
      begin
        N := BufToStr(M.LParam);
        if Options.PopupsOn then PopupManager.AddPopup('LEAVES!', N, POPUP_NORMAL);
        if Options.BeepOn then MessageBeep(DWORD(-1));
      end;
    TAKEVENT_PERSONAL:
      begin
        St := BufToStr(M.LParam);
        N := GetNextLine(St);
        C := FindUserInList(N);
        if (C <> -1) and not lbOnline.Items[C].Checked then C := -1;
        if (C = -1) and (N <> SockU.ChatName) then
        begin
          if Options.PopupsOn then PopupManager.AddPopup(N, St, POPUP_ERROR);
          if Options.BeepOn then MessageBeep(DWORD(-1));
        end;
      end;
    TAKEVENT_PRIVATE:
      begin
        St := BufToStr(M.LParam);
        N := GetNextLine(St);
        C := FindUserInList(N);
        if (C <> -1) and not lbOnline.Items[C].Checked then C := -1;
        if (C = -1) and (N <> SockU.ChatName) then
        begin
          if Options.PopupsOn then PopupManager.AddPopup(N, St, POPUP_WARNING);
          if Options.BeepOn then MessageBeep(DWORD(-1));
        end;
      end;
  end;
end;

procedure TfmMain.LogMsg(S: string);
var
  LS: string;
  I, St, F, E: Integer;
  C: DWORD;
  A: packed array[0..3] of Byte;
  A1: packed array[0..3] of Byte absolute C;
  FS: TFileStream;
  PrevA: Char;

  procedure AddText(const Str: string);
  var
    S: string;
  begin
    LS := Copy(Str, St, F - St);
    if (PrevA = #9) then
    begin
      //  
      Delete(LS, 1, 1);
      Delete(LS, Length(LS), 1);
      //   
      I := 0;
      while (I < lbOnline.Items.Count) and (lbOnline.Items[I].Caption <> LS) do Inc(I);
      if (I < lbOnline.Items.Count) then
      begin
        I := Integer(lbOnline.Items[I].Data); //  
      end
      else
        I := 0;
      View.AddActionString('[' + LS + ']', NameClick, I);
    end
    else
    begin
      while (LS <> '') do
      begin
        S := IsMail(LS);  // check for mailto:
        if (S <> '') then
        begin
          View.AddActionStringWithStyle(S,
            GlyphStyleRec([fsUnderline], clBlue, 'Tahoma', 10),
            MailClick, 0);
          Delete(LS, 1, Length(S));
        end;
        if (S = '') then
        begin  // check for URL
          S := IsURL(LS);
          if (S <> '') then
          begin
            if (S[1] = ' ') then
            begin
              View.AddString(LS[1]);
              Delete(S, 1, 1);
              Delete(LS, 1, 1);
            end;
            View.AddActionStringWithStyle(S,
              GlyphStyleRec([fsUnderline], clBlue, 'Tahoma', 10),
              URLClick, 0);
            Delete(LS, 1, Length(S));
          end;
        end;
        if (S = '') then
        begin  // normal text
          View.AddString(LS[1]);
          Delete(LS, 1, 1);
        end;
      end;
    end;
  end;

begin
  // save to log file
  LS := S + #10;
  FS := nil;
  try
    FS := TFileStream.Create('chat.log', fmOpenWrite or fmShareDenyWrite);
    FS.Position := FS.Size;
  except
    try
      FS := TFileStream.Create('chat.log', fmCreate);
    except
       FreeAndNil(FS);
    end;
  end;
  if (FS <> nil) then
  begin
    try
      if (LS <> '') then FS.WriteBuffer(LS[1], Length(LS));
    except
    end;
    FS.Free;
  end;

  //#0: normal attrs
  //#1-#2: picture
  //#3: private
  //#4: personal
  //#5: normal time
  //#6: HELLO time
  //#7: PERSONAL time
  //#8: PRIVATE time
  //#9: name
  //#31: set new default color

  View.OpenParagraph;
  try
    View.SetStyle(GlyphStyleRec([], clBlack, 'Tahoma', 10));

    F := 1;
    St := 1;
    C := 0;
    PrevA := #0;

    //  
    while (F <= Length(S)) do
    begin
      if (S[F] < #10) or (S[F] = #31) then
      begin
        AddText(S);

        PrevA := S[F];
        case PrevA of
          #0:  // normal attrs
            begin
              View.SetStyle(GlyphStyleRec([], C, 'Tahoma', 10));
              Inc(F);
            end;
          #1:  // picture
            begin
              Inc(F);
              E := F;
              while (E <= Length(S)) and (S[E] <> #2) do Inc(E);
              View.AddImage(Copy(S, F, E - F));
              F := E + 1;
            end;
          #2: ;
          #3:  // private
            begin
              View.SetStyle(GlyphStyleRec([fsItalic, fsBold], $0000FF, 'Tahoma', 10));
              Inc(F);
            end;
          #4:  // personal
            begin
              View.SetStyle(GlyphStyleRec([fsBold], $0000FF, 'Tahoma', 10));
              Inc(F);
            end;
          #5:  // normal time
            begin
              View.SetStyle(GlyphStyleRec([fsBold], 0, 'Tahoma', 10));
              Inc(F);
            end;
          #6:  // HELLO time
            begin
              View.SetStyle(GlyphStyleRec([fsBold], $FF0000, 'Tahoma', 10));
              Inc(F);
            end;
          #7:  // PERSONAL time
            begin
              View.SetStyle(GlyphStyleRec([fsItalic, fsBold], $0000FF, 'Tahoma', 10));
              Inc(F);
            end;
          #8:  // PRIVATE time
            begin
              View.SetStyle(GlyphStyleRec([fsBold], $0000FF, 'Tahoma', 10));
              Inc(F);
            end;
          #9:  // name
            begin
              View.SetStyle(GlyphStyleRec([fsUnderline, fsBold], $FF0000, 'Tahoma', 10));
              Inc(F);
            end;
          #31:  // set new default color
            begin
              Move(S[F + 1], A, 4);
              Inc(F, 5);
              A1[2] := A[0];
              A1[1] := A[1];
              A1[0] := A[2];
            end;
        end;
        St := F;
      end
      else
        Inc(F);
    end;
    AddText(S);  //    ?
  finally
    View.CloseParagraph;
  end;
end;

procedure TfmMain.DisableAll;
begin
  DisableAction(ActionEnter);
  DisableAction(ActionLeave);
  DisableAction(ActionRefresh);
  DisableAction(ActionSay);
end;

procedure TfmMain.EnableAll;
begin
  EnableAction(ActionEnter);
  EnableAction(ActionLeave);
  EnableAction(ActionRefresh);
  EnableAction(ActionSay);
end;

procedure TfmMain.DoLogin;
begin
  if (RThread <> nil) then
  begin
    if Options.BeepOn then MessageBeep(DWORD(-1));
    exit;
  end;
  DisableAll;

  ActionLeave.Tag := 0;
  ActionRefresh.Tag := 0;
  ActionSay.Tag := 0;

  if Options.UseProxy then
  begin
    SockU.ProxyAddr := Options.ProxyAddr + ':' + IntToStr(Options.ProxyPort);
    if Options.UseAuth then
      SockU.ProxyAuth := Options.AuthUser + ':' + Options.AuthPassword
    else
      SockU.ProxyAuth := '';
  end
  else
  begin
    SockU.ProxyAddr := '';
    SockU.ProxyAuth := '';
  end;
  SockU.ChatName := Options.LoginName;

  if Options.RememberIds then
    LoadIds
  else
  begin
    SockU.ChatName := '';
    SockU.ChatId := '';
    SockU.ChatLSize := '';
    SockU.ChatLId := '';
  end;
  SockU.MastakServer := Options.MasterServer;

  RThread := TRefreshThread.Create(Handle, Options.LoginName,
    Options.LoginPassword, Options.ChatRefTime, Options.UListRefTime);
end;

procedure TfmMain.DoLogout;
begin
  if (RThread = nil) then
  begin
    if Options.BeepOn then MessageBeep(DWORD(-1));
    exit;
  end;

  DisableAll;
  RThread.Terminate;
  while (RThread <> nil) do
  begin
    Application.ProcessMessages;
    Sleep(100);
  end;

  SockU.DoLogout;

  ActionEnter.Tag := 1;
  ActionLeave.Tag := 0;
  ActionRefresh.Tag := 0;
  ActionSay.Tag := 0;
  EnableAll;
end;

procedure TfmMain.lbOnlineDblClick(Sender: TObject);
var
  S: string;
begin
  if (lbOnline.Selected = nil) then exit;

  if (GetAsyncKeyState(VK_SHIFT) and $8000 <> 0) then
  begin
    ShellExecute(Application.Handle, 'open',
      PChar('http://' + SockU.MastakServer + '/cgi-bin/anketa.pl?id=' + IntToStr(Integer(lbOnline.Selected.Data))),
      '', '', SW_SHOWNORMAL);
  end
  else
  begin
    if (GetAsyncKeyState(VK_CONTROL) and $8000 <> 0) then
      S := 'private'
    else
      S := 'to';
    cbText.Text := S + ' [' + lbOnline.Selected.Caption + '] ' + cbText.Text;
  end;
  lbOnline.Selected := nil;
  ActivateEdit;
end;

procedure TfmMain.ActivateEdit;
begin
  ActiveControl := cbText;
  cbText.SelectAll;
  cbText.SelStart := -1;
end;

procedure TfmMain.FormCreate(Sender: TObject);
begin
  Application.Title := TAK_VERSION;
  Caption := TAK_VERSION + ' by Ketmar // Avalon Group';

  TrayIconActive := False;

  TrayIconData.cbSize := SizeOf(TNotifyIconData);
  TrayIconData.Wnd := Handle;
  TrayIconData.uId := 1;
  TrayIconData.uFlags := NIF_ICON or NIF_TIP or NIF_MESSAGE;
  TrayIconData.uCallBackMessage := WM_TRAYMSG;
  TrayIconData.hIcon := LoadIcon(HInstance, 'MAINICON');
  FillChar(TrayIconData.szTip, 64, 0);
  TrayIconData.szTip := 'TAKChat!';

  View := TTAKView.Create(Self);
  View.Parent := Self;
  View.Align := alClient;

  RThread := nil;
  LoadSmileFile(ExtractFilePath(ParamStr(0)) + 'smiles.rc');

  if not LoadIni then PostMessage(Handle, WM_USER, 0, 0);
  PopupManager := TPopupManager.Create(Options.PopupsTop, Options.PopupsLeft, 5000);
end;

procedure TfmMain.AppEventsMinimize(Sender: TObject);
begin
  if Options.TrayIcon then
  begin
    SetTrayIconStatus(True);
    ShowWindow(Application.Handle, SW_HIDE);
  end;
end;

procedure TfmMain.AppEventsRestore(Sender: TObject);
begin
  ShowWindow(Application.Handle, SW_NORMAL);
  SetForegroundWindow(Handle);
  View.ToBottom;
  ActivateEdit;
  if Options.TrayIcon then SetTrayIconStatus(Options.PermanentIcon);
end;

procedure TfmMain.SetDefaultOptions;
begin
  Options.UseProxy := False;
  Options.ProxyAddr := '';
  Options.ProxyPort := 8080;
  Options.UseAuth := False;
  Options.AuthUser := '';
  Options.AuthPassword := '';
  Options.LoginName := '';
  Options.LoginPassword := '';
  Options.TrayIcon := True;
  Options.PermanentIcon := True;
  Options.BeepOn := True;
  Options.PopupsOn := True;
  Options.PopupsLeft := True;
  Options.PopupsTop := True;
  Options.DebugLog := False;
  Options.RememberIds := False;
  Options.ChatRefTime := 15;
  Options.UListRefTime := 2;
  Options.MasterServer := 'www.delphimaster.ru';
  CurColor := 0;
  LogU.DoLog := Options.DebugLog;
end;

function TfmMain.LoadIni: Boolean;
var
  IniF: TStream;
  Pos: TRect;
  W: Integer;
  Ver: Byte;
begin
  Result := False;
  SetDefaultOptions;

  try
    IniF := TFileStream.Create(ExtractFilePath(ParamStr(0)) + 'takchat.opt', fmOpenRead or fmShareDenyWrite);
  except
    LogStr('can''t open ini file');
    exit;
  end;

  try
    try
      IniF.ReadBuffer(Ver, SizeOf(Ver));
      if (Ver <> 1) and (Ver <> 2) then exit;
      IniF.ReadBuffer(CurColor, SizeOf(CurColor)); SetColor;
      Options.UseProxy := LoadBool(IniF);
      Options.ProxyAddr := LoadStr(IniF);
      IniF.ReadBuffer(Options.ProxyPort, SizeOf(Options.ProxyPort));
      Options.UseAuth := LoadBool(IniF);
      Options.AuthUser := LoadStr(IniF);
      Options.AuthPassword := LoadStr(IniF);
      Options.LoginName := LoadStr(IniF);
      Options.LoginPassword := LoadStr(IniF);
      Options.TrayIcon := LoadBool(IniF);
      Options.PermanentIcon := LoadBool(IniF);
      Options.BeepOn := LoadBool(IniF);
      Options.PopupsOn := LoadBool(IniF);
      Options.PopupsLeft := LoadBool(IniF);
      Options.PopupsTop := LoadBool(IniF);
      Options.DebugLog := LoadBool(IniF);
      Options.RememberIds := LoadBool(IniF);
      Options.ChatRefTime := LoadInt(IniF);
      Options.UListRefTime := LoadInt(IniF);

      Pos.Left := LoadInt(IniF);
      Pos.Top := LoadInt(IniF);
      Pos.Right := LoadInt(IniF);
      Pos.Bottom := LoadInt(IniF);
      SetBounds(Pos.Left, Pos.Top, Pos.Right, Pos.Bottom);
      W := LoadInt(IniF);;
      lbOnline.Width := W;
      LogU.DoLog := Options.DebugLog;
      Result := True;

      if Ver > 1 then Options.MasterServer := LoadStr(IniF);
    except
      LogStr('error reading ini file');
    end;
  finally
    if not Result then SetDefaultOptions;
    IniF.Free;
  end;
end;

procedure TfmMain.SaveIni;
var
  IniF: TStream;
  Ver: Byte;
begin
  try
    IniF := TFileStream.Create(ExtractFilePath(ParamStr(0)) + 'takchat.opt', fmCreate);
  except
    LogStr('can''t create ini file');
    exit;
  end;

  try
    try
      Ver := 1;
      IniF.WriteBuffer(Ver, SizeOf(Ver));
      IniF.WriteBuffer(CurColor, SizeOf(CurColor)); SetColor;
      SaveBool(IniF, Options.UseProxy);
      SaveStr(IniF, Options.ProxyAddr);
      IniF.WriteBuffer(Options.ProxyPort, SizeOf(Options.ProxyPort));
      SaveBool(IniF, Options.UseAuth);
      SaveStr(IniF, Options.AuthUser);
      SaveStr(IniF, Options.AuthPassword);
      SaveStr(IniF, Options.LoginName);
      SaveStr(IniF, Options.LoginPassword);
      SaveBool(IniF, Options.TrayIcon);
      SaveBool(IniF, Options.PermanentIcon);
      SaveBool(IniF, Options.BeepOn);
      SaveBool(IniF, Options.PopupsOn);
      SaveBool(IniF, Options.PopupsLeft);
      SaveBool(IniF, Options.PopupsTop);
      SaveBool(IniF, Options.DebugLog);
      SaveBool(IniF, Options.RememberIds);
      SaveInt(IniF, Options.ChatRefTime);
      SaveInt(IniF, Options.UListRefTime);

      SaveInt(IniF, Left);
      SaveInt(IniF, Top);
      SaveInt(IniF, Width);
      SaveInt(IniF, Height);
      SaveInt(IniF, lbOnline.Width);

      SaveStr(IniF, Options.MasterServer);
    except
      LogStr('error reading ini file');
    end;
  finally
    IniF.Free;
  end;
end;

procedure TfmMain.RxTrayIconDblClick(Sender: TObject);
begin
  Application.Restore;
end;

procedure TfmMain.Panel2Resize(Sender: TObject);
begin
  cbText.Width := Panel2.Width - 68;
end;

procedure TfmMain.FormDestroy(Sender: TObject);
begin
  if (RThread <> nil) then
  begin
    RThread.Terminate;
    TerminateThread(RThread.Handle, 0);
    RThread := nil;
  end;
//  FreeAndNil(PopupManager);
  SetTrayIconStatus(False);
  SaveIni;
end;

procedure TfmMain.LoadIds;
var
  FO: TFileStream;
begin
  try
    FO := TFileStream.Create(ExtractFilePath(ParamStr(0)) + 'ids.bin', fmOpenRead or fmShareDenyWrite);
    try
      SockU.ChatId := LoadStr(FO);
      SockU.ChatLSize := LoadStr(FO);
      SockU.ChatLId := LoadStr(FO);
      SockU.ChatName := Options.LoginName;
    finally
      FO.Free;
    end;
  except
    SockU.ChatId := '';
    SockU.ChatLSize := '';
    SockU.ChatLId := '';
  end;
end;

procedure TfmMain.SaveIds;
var
  FO: TFileStream;
begin
  try
    FO := TFileStream.Create(ExtractFilePath(ParamStr(0)) + 'ids.bin', fmCreate);
    try
      SaveStr(FO, SockU.ChatId);
      SaveStr(FO, SockU.ChatLSize);
      SaveStr(FO, SockU.ChatLId);
    finally
      FO.Free;
    end;
  except
    DeleteFile(ExtractFilePath(ParamStr(0)) + 'ids.bin');
  end;
end;

procedure TfmMain.SetColor;
var
  S: string;
begin
  S := IntToHex(CurColor, 6);
  //123456
  ChatColor := S[5] + S[6] + S[3] + S[4] + S[1] + S[2];
end;

procedure TfmMain.NameClick(Sender: TGlyph; ReaderPos: TReaderProsition);
var
  S: string;
  Reader: TTAKReader;
  Name: string;
  N: Integer;
begin
  Reader := View.Text.NewReader;
  try
    Reader.SetPosition(ReaderPos);
    Name := Reader.GetUnderlinedText;
  finally
    Reader.Free;
  end;
  Delete(Name, 1, 1);
  Delete(Name, Length(Name), 1);
  if (Name = '') then exit;

  if (GetAsyncKeyState(VK_SHIFT) and $8000 <> 0) then
  begin
    N := Sender.UserData;
    if (N = 0) then
    begin
      N := FindUserInList(Name);
      if (N < 0) then N := 0 else N := Integer(lbOnline.Items[N].Data);
    end;

    if (N <> 0) then
      ShellExecute(Application.Handle, 'open',
        PChar('http://' + SockU.MastakServer + '/cgi-bin/anketa.pl?id=' + IntToStr(N)),
        '', '', SW_SHOWNOACTIVATE);
  end
  else
  begin
    if (GetAsyncKeyState(VK_CONTROL) and $8000 <> 0) then
      S := 'private'
    else
      S := 'to';
    cbText.Text := S + ' [' + Name + '] ' + cbText.Text;
  end;

  ActivateEdit;
end;

procedure TfmMain.cbTextKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = VK_ESCAPE) then
  begin
    cbText.Text := '';
    Key := 0;
  end;
end;

procedure TfmMain.SetTrayIconStatus(Active: Boolean);
begin
  if (TrayIconActive = Active) then exit;
  if (TrayIconActive = Active) then exit;
  TrayIconActive := Active;
  if Active then
  begin
    if not Shell_NotifyIcon(NIM_ADD, @TrayIconData) then
      Shell_NotifyIcon(NIM_MODIFY, @TrayIconData);
  end
  else
    Shell_NotifyIcon(NIM_DELETE, @TrayIconData);
end;

procedure TfmMain.MsgTray(var M: TMessage);
begin
  if (M.lParam = WM_LBUTTONDBLCLK) then RxTrayIconDblClick(nil);
end;

procedure TfmMain.MailClick(Sender: TGlyph; ReaderPos: TReaderProsition);
var
  Reader: TTAKReader;
  Adr: string;
begin
  Reader := View.Text.NewReader;
  try
    Reader.SetPosition(ReaderPos);
    Adr := Reader.GetUnderlinedText;
  finally
    Reader.Free;
  end;
  if (Adr = '') then exit;

  if (Copy(Adr, 1, 7) <> 'mailto:') then Adr := 'mailto:' + Adr;
  ShellExecute(Application.Handle, 'open', PChar(Adr), '', '', SW_SHOWNORMAL);

  ActivateEdit;
end;

procedure TfmMain.URLClick(Sender: TGlyph; ReaderPos: TReaderProsition);
var
  Reader: TTAKReader;
  Adr: string;
begin
  Reader := View.Text.NewReader;
  try
    Reader.SetPosition(ReaderPos);
    Adr := Reader.GetUnderlinedText;
  finally
    Reader.Free;
  end;
  if (Adr = '') then exit;
  if (Copy(Adr, 1, 7) <> 'http://') and (Copy(Adr, 1, 6) <> 'ftp://') then
    Adr := 'http://' + Adr;

  ShellExecute(Application.Handle, 'open', PChar(Adr), '', '', SW_SHOWNORMAL);

  ActivateEdit;
end;

function GetTillChar(const S: string; const Ch: Char): string;
var
  P: Integer;
begin
  Result := '';
  P := 1;
  while (P <= Length(S)) and (S[P] > ' ') and (S[P] <> Ch) do Inc(P);
  if (P <= Length(S)) and (S[P] = Ch) then Result := Copy(S, 1, P - 1);
end;

function GetText(const S: string): string;
var
  P: Integer;
begin
  Result := '';
  P := 1;
  while (P <= Length(S)) and (S[P] > ' ') do Inc(P);
  Result := Copy(S, 1, P - 1);
end;

function TfmMain.IsMail(S: string): string;
begin
  Result := GetTillChar(S, ':');
  if (Result = 'mailto') then
  begin
    Result := GetText(S);
    exit;
  end;

  Result := GetTillChar(S, '@');
  if (Result <> '') then Result := GetText(S);
end;

function TfmMain.IsURL(S: string): string;
var
  P: Integer;
begin
  Result := GetText(S);
  if (Copy(Result, 1, 7) = 'http://') then exit;
  if (Copy(Result, 1, 6) = 'ftp://') then exit;
  if (Copy(Result, 1, 4) = 'www.') then exit;

  //   .org, .net...
  Result := GetText(Copy(S, 2, Length(S)));
  if (Copy(S, 1, 1) <= ' ') and (Result <> '') and (Pos('@', Result) = 0) then
  begin
    Result := ' ' + Result;
    P := Pos('/', Result);
    if (P = 0) then P := Length(Result) + 1;
    S := Copy(Result, P - 3, 3);
    if (S = '.tv') or (S = '.fm') or (S = '.ru') or (S = '.ua') or (S = '.tk') then exit;
    S := Copy(Result, P - 4, 4);
    if (S = '.net') or (S = '.com') or (S = '.org') or (S = '.gov') or (S = '.edu') or
       (S = '.biz') then exit;
    S := Copy(Result, P - 5, 5);
    if (S = '.info') then exit;
  end;
  Result := '';
end;

function TfmMain.FindUserInList(Name: string): Integer;
{var
  P: Integer;}
begin
{  if (Name <> '') and (Name[1] = '[') then
  begin
    Delete(Name, 1, 1);
    P := 1;
    while (P <= Length(Name)) and (Name[P] <> ']') do Inc(P);
    Name := Copy(Name, 1, P - 1);
  end;}
  if (Name <> '') then
  begin
    Result := 0;
    while (Result < lbOnline.Items.Count) and
          (ANSICompareText(Name, lbOnline.Items[Result].Caption) <> 0) do Inc(Result);
    if (Result < lbOnline.Items.Count) then exit;
  end;

  Result := -1;
end;

procedure TfmMain.AddToHistory(const S: string);
var
  F: Integer;
begin
  if (S <> '') then
  begin
    F := 0;
    while (F < cbText.Items.Count) and (cbText.Items[F] <> S) do Inc(F);
    if (F = cbText.Items.Count) then
    begin
      if (cbText.Items.Count = 20) then cbText.Items.Delete(19);
    end
    else
      cbText.Items.Delete(F);
    cbText.Items.Insert(0, S);
  end;
end;

procedure TfmMain.ActionEnterExecute(Sender: TObject);
begin
  DoLogin;
  ActivateEdit;
end;

procedure TfmMain.ActionLeaveExecute(Sender: TObject);
begin
  DoLogout;
  Close;
end;

procedure TfmMain.ActionRefreshExecute(Sender: TObject);
begin
  if (RThread <> nil) then
  begin
    RThread.ForceOnlineRefresh;
    RThread.ForceChatRefresh;
  end;
end;

procedure TfmMain.ActionSetColorExecute(Sender: TObject);
begin
  ColorDialog.Color := CurColor;
  if ColorDialog.Execute then
  begin
    CurColor := ColorDialog.Color;
    SetColor;
    SaveIni;
  end;
end;

procedure TfmMain.ActionSayExecute(Sender: TObject);
begin
  if (RThread <> nil) then
  begin
    LoadSmileFile(ExtractFilePath(ParamStr(0)) + 'smiles.rc');
    RThread.Say(ReplaceSmiles(cbText.Text));
    AddToHistory(cbText.Text);
    cbText.Text := '';
    ActivateEdit;
  end;
end;

procedure TfmMain.miExitClick(Sender: TObject);
begin
  Close;
end;

procedure TfmMain.ActionPrefsExecute(Sender: TObject);

  function NormSrv (const s: string): string;
  begin
    result := ANSILowerCase(s);
    if Copy(result, 1, 7) = 'http://' then Delete(result, 1, 7);
    if (result <> '') and (result[Length(result)] in ['/', '\']) then Delete(result, Length(result), 1);
  end;

var
  F: TfmPrefs;
begin
  F := TfmPrefs.Create(Application);
  try
    F.edMasterServer.Text := Options.MasterServer;
    F.cbUseProxy.Checked := Options.UseProxy;
    F.edProxyAddr.Text := Options.ProxyAddr;
    F.seProxyPort.Value := Options.ProxyPort;
    F.cbProxyAuth.Checked := Options.UseAuth;
    F.edProxyUser.Text := Options.AuthUser;
    F.edProxyPass.Text := Options.AuthPassword;
    F.edLogin.Text := Options.LoginName;
    F.edPassword.Text := Options.LoginPassword;
    F.cbTrayIcon.Checked := Options.TrayIcon;
    F.cbPermanentTray.Checked := Options.PermanentIcon;
    F.cbDebugLog.Checked := Options.DebugLog;
    F.cbRememberIds.Checked := Options.RememberIds;
    F.cbBeepOn.Checked := Options.BeepOn;
    F.seChat.Value := Options.ChatRefTime;
    F.seUList.Value := Options.UListRefTime;
    if Options.PopupsOn then
    begin
      if Options.PopupsLeft then
      begin
        if Options.PopupsTop then
          F.rgPopupPos.ItemIndex := 1
        else
          F.rgPopupPos.ItemIndex := 2;
      end
      else
      begin
        if Options.PopupsTop then
          F.rgPopupPos.ItemIndex := 3
        else
          F.rgPopupPos.ItemIndex := 4;
      end;
    end
    else
      F.rgPopupPos.ItemIndex := 0;

    if (F.ShowModal = mrOK) then
    begin
      Options.MasterServer := NormSrv(F.edMasterServer.Text);
      Options.UseProxy := F.cbUseProxy.Checked;
      Options.ProxyAddr := F.edProxyAddr.Text;
      Options.ProxyPort := F.seProxyPort.Value;
      Options.UseAuth := F.cbProxyAuth.Checked;
      Options.AuthUser := F.edProxyUser.Text;
      Options.AuthPassword := F.edProxyPass.Text;
      Options.LoginName := F.edLogin.Text;
      Options.LoginPassword := F.edPassword.Text;
      Options.TrayIcon := F.cbTrayIcon.Checked;
      Options.PermanentIcon := F.cbPermanentTray.Checked;
      Options.DebugLog := F.cbDebugLog.Checked;
      Options.RememberIds := F.cbRememberIds.Checked;
      Options.BeepOn := F.cbBeepOn.Checked;
      Options.ChatRefTime := F.seChat.Value;
      Options.UListRefTime := F.seUList.Value;

      if (RThread <> nil) then
      begin
        RThread.SetNewIntervals(Options.ChatRefTime, Options.UListRefTime);
      end;

      Options.PopupsOn := (F.rgPopupPos.ItemIndex > 0);
      Options.PopupsLeft := (F.rgPopupPos.ItemIndex = 1) or (F.rgPopupPos.ItemIndex = 2);
      Options.PopupsTop := (F.rgPopupPos.ItemIndex = 1) or (F.rgPopupPos.ItemIndex = 3);
      PopupManager.FromTop := Options.PopupsTop;
      PopupManager.FromLeft := Options.PopupsLeft;

      LogU.DoLog := Options.DebugLog;

      SaveIni;
    end;
  finally
    F.Free;
  end;
end;

procedure TfmMain.MsgUser(var M: TMessage);
begin
  ActionPrefs.Execute;
end;

procedure TfmMain.ActionCutLogExecute(Sender: TObject);
begin
  while (View.Text.ParagraphCount > 30) do View.Text.DeleteParagraph(0);
  View.Reformat;
end;

end.
